// +-------------------------------------------------------------------------+
// | FlashQard - An efficient way to memorize.                               |
// | Copyright (C) 2008-2009 Shahab <shahab@flashqard-project.org>           |
// +-------------------------------------------------------------------------+
// | License:                                                                |
// |                                                                         |
// | FlashQard is free software; you can redistribute it and/or modify       |
// | it under the terms of the GNU General Public License as published by    |
// | the Free Software Foundation; either version 2 of the License, or       |
// | (at your option) any later version.                                     |
// |                                                                         |
// | FlashQard is distributed in the hope that it will be useful,            |
// | but WITHOUT ANY WARRANTY; without even the implied warranty of          |
// | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the            |
// | GNU General Public License for more details.                            |
// |                                                                         |
// | You should have received a copy of the GNU General Public License       |
// | along with FlashQard; if not, write to the Free Software                |
// | Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA|
// |                                                                         |
// +-------------------------------------------------------------------------+
// | Author:  Shahab Shirazi <shahab@flashqard-project.org>                  |
// +-------------------------------------------------------------------------+

#include <QtGui>
#include <QPainterPath>
#include "CardBoxTreeWidget_Stack.h"
#include "Settings.h"
#include "SearchConfigureDialog.h"
#include "Utilities.h"
#include "MimeData.h"

TreeWidgetForStack::TreeWidgetForStack(CardBoxTreeWidget *parent)
   : TreeWidget(parent),
     cardBoxTreeWidget(parent)
{
   
}

void TreeWidgetForStack::dropEvent(QDropEvent *event)
{
   QTreeWidgetItem *item = itemAt(event->pos());
   if (!item)
      return;
   const MimeData *mime = dynamic_cast<const MimeData*>(event->mimeData());
   if (!mime)
      return;
   cardBoxTreeWidget -> dropMimeData(item, mime, event->dropAction());
}

QMimeData* TreeWidgetForStack::mimeData(const QList<QTreeWidgetItem*> items) const
{
   return cardBoxTreeWidget -> mimeData(items);
}

QStringList TreeWidgetForStack::mimeTypes() const
{
   return cardBoxTreeWidget -> mimeTypes();
}

Qt::DropActions TreeWidgetForStack::supportedDropActions() const
{
   return Qt::MoveAction;
}

//======================================================================

CardBoxTreeWidget_Stack::CardBoxTreeWidget_Stack(QWidget *parent)
   : CardBoxTreeWidget(parent)
{
   searchConfigureDialog = new SearchConfigureDialog(this);

   treeWidget = new TreeWidgetForStack(this);
   treeWidget->setSelectionMode(QAbstractItemView::ExtendedSelection);
   treeWidget->setDragDropMode(QAbstractItemView::DragDrop);
   treeWidget->setColumnCount(1);
//   treeWidget->resizeColumnToContents(0);
   QStringList headerLabels;
   headerLabels << tr("Title");
   treeWidget->setHeaderLabels(headerLabels);

   searchLabel = new QLabel(tr("&Search"), this);
   searchLine = new QLineEdit(this);
   searchConfigureButton = new QToolButton(this);
   searchConfigureButton->setText(tr("..."));
   searchLabel->setBuddy(searchLine);
   QHBoxLayout *horizontalLayout = new QHBoxLayout();
   horizontalLayout->addWidget(searchLabel);
   horizontalLayout->addWidget(searchLine);
   horizontalLayout->addWidget(searchConfigureButton);
   QVBoxLayout *verticalLayout = new QVBoxLayout();
   verticalLayout->setContentsMargins(0,0,0,0);
   verticalLayout->addLayout(horizontalLayout);
   verticalLayout->addWidget(treeWidget);
   setLayout(verticalLayout);

   connect (treeWidget, SIGNAL(itemActivated(QTreeWidgetItem*, int)), this, SLOT(itemActivated(QTreeWidgetItem*)));
   connect (treeWidget, SIGNAL(currentItemChanged()), this, SIGNAL(currentItemChanged()));
   connect (searchLine, SIGNAL(textChanged(const QString&)), this, SLOT(updateSearch()));
   connect (searchConfigureButton, SIGNAL(clicked()), this, SLOT(searchConfigureButtonClicked()));
}

QSize CardBoxTreeWidget_Stack::sizeHint() const
{
   int width = treeWidget->columnWidth(0) + 15;
   return QSize(width, size().height());
}

QList<CardPosition> CardBoxTreeWidget_Stack::selectedPositions() const
{
   QList<QTreeWidgetItem*> selectedItems = treeWidget -> selectedItems();
   QList<CardPosition> selectedPositions;

   for (int i=0; i<selectedItems.size(); i++)
   {
      CardPosition position;
      cardPositionOfItem(selectedItems[i], &position);
      selectedPositions << position;
   }
   return selectedPositions;
}

ItemType CardBoxTreeWidget_Stack::currentCardPosition(CardPosition *position) const
{
   return cardPositionOfItem(treeWidget->currentItem(), position);
}

bool CardBoxTreeWidget_Stack::isAnyCardSelected() const
{
   QList<QTreeWidgetItem*> selectedItems = treeWidget -> selectedItems();

   for (int i=0; i<selectedItems.size(); i++)
   {
      CardPosition position;
      if (cardPositionOfItem(selectedItems[i], &position) == ItemType_Card)
         return true;
   }
   
   return false;
}

bool CardBoxTreeWidget_Stack::isAnythingSelected() const
{
   return treeWidget->selectedItems().size();
}

void CardBoxTreeWidget_Stack::selectCard(const CardPosition &pos)
{
   if (pos.cardNumber < -1 || pos.cardNumber >= cardBox->countStack())
      return;
   QTreeWidgetItem *topLevelItem = treeWidget->topLevelItem(0);
   if (pos.cardNumber == -1)
   {
      treeWidget->setCurrentItem(topLevelItem);
      return;
   }
   else if (!topLevelItem->child(pos.cardNumber)->isHidden())
      treeWidget->setCurrentItem(topLevelItem->child(pos.cardNumber));
}

void CardBoxTreeWidget_Stack::setSelectionMode(QAbstractItemView::SelectionMode mode)
{
   treeWidget -> setSelectionMode(mode);
}

QMimeData* CardBoxTreeWidget_Stack::mimeData(const QList<QTreeWidgetItem*> items) const
{
   MimeData *mime = new MimeData;
   for (int i=0; i<items.size(); i++)
   {
      CardPosition pos;
      if (cardPositionOfItem(items[i], &pos) != ItemType_None)
         mime -> addCardPosition(pos);
   }
   
   return mime;
}

void CardBoxTreeWidget_Stack::dropMimeData(QTreeWidgetItem *item, const MimeData *mime, Qt::DropAction action)
{
   CardPosition insertPos;
   cardPositionOfItem(item, &insertPos);

   emit dropMimeDataSignal(insertPos, mime, action);
}

void CardBoxTreeWidget_Stack::setDragDropMode(QAbstractItemView::DragDropMode mode)
{
   treeWidget -> setDragDropMode(mode);
}

QAbstractItemView::DragDropMode CardBoxTreeWidget_Stack::dragDropMode() const
{
   return treeWidget -> dragDropMode();
}

void CardBoxTreeWidget_Stack::updateCards()
{
   QDate currentDate = QDate::currentDate();
   bool isExpanded = false;
   if (treeWidget->topLevelItemCount())
      isExpanded = treeWidget->topLevelItem(0)->isExpanded();

   treeWidget -> clear();
   QTreeWidgetItem *parent = new QTreeWidgetItem(treeWidget, QStringList(tr("Stack   (%2)")
                                                                         .arg(cardBox->countStack())));
   for (int column=0; column<treeWidget->columnCount(); column++)
      parent -> setBackground(column, QBrush(Qt::gray));
   for (int i = 0; i<cardBox->countStack(); i++)
   {
      
      CardPosition pos(i);
      QString title;
      title = cardBox->at(pos)->getTitle();
      QTreeWidgetItem *child = new QTreeWidgetItem(parent, QStringList(title));

      QColor iconColor;
      if (cardBox->at(pos) -> getCardStatistics().getDateCreated() == currentDate)
         child -> setIcon(0, createIcon(Settings::settings() -> generalNewCards));
   }

   if (!searchLine->text().isEmpty())
      updateSearch();

   if (isExpanded)
      treeWidget->topLevelItem(0)->setExpanded(true);
}

void CardBoxTreeWidget_Stack::updateCurrentCardTitle()
{
   QTreeWidgetItem *item = treeWidget->currentItem();
   CardPosition pos;
   if (cardPositionOfItem(item, &pos) == ItemType_Card)
      item -> setText(0, cardBox->at(pos)->getTitle());
}

void CardBoxTreeWidget_Stack::itemActivated(QTreeWidgetItem *item)
{
   CardPosition pos;
   cardPositionOfItem(item, &pos);
   emit cardActivated(pos);
}

void CardBoxTreeWidget_Stack::updateSearch()
{
   QString searchString = searchLine->text();
   if (searchString.isEmpty())
   {
      updateCards();
      return;
   }
   setAllItemsHidden(true);
   FindIn findIn = searchConfigureDialog->getFindIn();
   bool isRegExp = searchConfigureDialog->isRegExp();

   if (findIn == FindIn_Title)
   {
      for (int i=0; i<cardBox->countStack(); i++)
      {
         QTreeWidgetItem *item = treeWidget->topLevelItem(0)->child(i);
         bool contains;
         if (isRegExp)
            contains = item->text(0).contains(QRegExp(searchString));
         else
            contains = item->text(0).contains(QRegExp(searchString, Qt::CaseInsensitive, QRegExp::Wildcard));
         if (contains)
            item->setHidden(false);
      }
      return;
   }
   
   CardPosition searchPosition(-1);
   while(1)
   {
      searchPosition = cardBox->search(searchString, searchPosition, isRegExp);
      if (!cardBox->isPositionValid(searchPosition))
         break;
      treeWidget->topLevelItem(0)->child(searchPosition.cardNumber)->setHidden(false);
      //stay responsive:
      qApp->processEvents();
      if (searchLine->text() != searchString)
         break;
   }
}

void CardBoxTreeWidget_Stack::searchConfigureButtonClicked()
{
   searchConfigureDialog->exec();
   updateSearch();
}

ItemType CardBoxTreeWidget_Stack::cardPositionOfItem(QTreeWidgetItem *item, CardPosition *position) const
{
   if (!item || !position)
      return ItemType_None;
   QTreeWidgetItem *myItem = static_cast<QTreeWidgetItem*>(item);
   position -> boxSection = BoxSection_Stack;
   if (myItem -> parent())
      position -> cardNumber = myItem->parent()->indexOfChild(myItem);
   else
      position -> cardNumber = -1;

   if (position -> cardNumber == -1)
      return ItemType_Stack;
   else
      return ItemType_Card;
}

void CardBoxTreeWidget_Stack::setAllItemsHidden(bool hidden)
{
   for (int i=0; i<cardBox->countStack(); i++)
      treeWidget->topLevelItem(0)->child(i)->setHidden(hidden);
}


