// +-------------------------------------------------------------------------+
// | FlashQard - An efficient way to memorize.                               |
// | Copyright (C) 2008-2009 Shahab <shahab@flashqard-project.org>           |
// +-------------------------------------------------------------------------+
// | License:                                                                |
// |                                                                         |
// | FlashQard is free software; you can redistribute it and/or modify       |
// | it under the terms of the GNU General Public License as published by    |
// | the Free Software Foundation; either version 2 of the License, or       |
// | (at your option) any later version.                                     |
// |                                                                         |
// | FlashQard is distributed in the hope that it will be useful,            |
// | but WITHOUT ANY WARRANTY; without even the implied warranty of          |
// | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the            |
// | GNU General Public License for more details.                            |
// |                                                                         |
// | You should have received a copy of the GNU General Public License       |
// | along with FlashQard; if not, write to the Free Software                |
// | Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA|
// |                                                                         |
// +-------------------------------------------------------------------------+
// | Author:  Shahab Shirazi <shahab@flashqard-project.org>                  |
// +-------------------------------------------------------------------------+

#include "CardReviewerVocabulary.h"
#include "Card.h"
#include "CardVocabulary.h"
#include "PrettyButton.h"
#include "Settings.h"
#include "Utilities.h"
#include <QToolBar>

CardReviewerVocabulary::CardReviewerVocabulary(QWidget *parent)
   : CardReviewer(parent)
{
   setupUi(this);
   textViewer -> setReadOnly(true);
   readingID = 0;
   isFrontSideVisible = true;
   commentsTextViewer -> setVisible(false);
   synonymsListWidget -> setVisible(false);
   antonymsListWidget -> setVisible(false);
   splitter -> setVisible(false);

   languagesToolBar = new QToolBar(this);
   QWidget *widget = new QWidget(this);
   QVBoxLayout *widgetLayout = new QVBoxLayout(widget);
   widgetLayout -> addWidget(textViewer);
   widgetLayout -> addWidget(languagesToolBar);
   mainSplitter -> insertWidget(0, widget);
   
   mainSplitter -> setStretchFactor(0, 10);
   mainSplitter -> setStretchFactor(1, 1);
   connect(Reader::reader(), SIGNAL(readingFinished(int)), this, SLOT(readingFinishedSlot(int)));
}

void CardReviewerVocabulary::showCard(const Card *c)
{
   card = dynamic_cast<const CardVocabulary*>(c);
   if (!card)
   {
      qDebug("CardReviewerVocabulary::showCard(): received card is not a CardVocabulary");
      return;
   }

   frontSideLanguageIndex = -1;
   backSideLanguageIndex = -1;

   if (Settings::settings() -> reviewVocabularyCustomizedLanguage)
   {
      for (int i=0; i<card->getTranslationDocuments().size(); i++)
      {
         if (card->getTranslationDocuments()[i] -> getLanguageName()
             .compare(Settings::settings() -> reviewVocabularyFromLanguage, Qt::CaseInsensitive) == 0)
            frontSideLanguageIndex = i;
         else if (card->getTranslationDocuments()[i] -> getLanguageName()
                  .compare(Settings::settings() -> reviewVocabularyToLanguage, Qt::CaseInsensitive) == 0)
            backSideLanguageIndex = i;
      }
   }
   if (frontSideLanguageIndex == -1 && backSideLanguageIndex == -1)
   {
      frontSideLanguageIndex = 0;
      backSideLanguageIndex = 1;
   }
   else if (frontSideLanguageIndex == -1)
      frontSideLanguageIndex = backSideLanguageIndex == 0 ? 1 : backSideLanguageIndex-1;
   else if (backSideLanguageIndex == -1)
      backSideLanguageIndex = frontSideLanguageIndex == card->getTranslationDocuments().size()-1
         ? frontSideLanguageIndex-1 : frontSideLanguageIndex+1;
   
   createLanguageButtons();
   showFrontSide();
   updateHints();
}

void CardReviewerVocabulary::showFrontSide()
{
   isFrontSideVisible = true;
   selectLanguage(frontSideLanguageIndex);
}

void CardReviewerVocabulary::verify()
{
   if (isFrontSideVisible)
   {
      isFrontSideVisible = false;
      selectLanguage(backSideLanguageIndex);
   }
   else
      showFrontSide();
}


void CardReviewerVocabulary::updateHints()
{
   QString label = card -> wordTypeToString(tr(", "));
   if (label.isEmpty())
      wordTypeLabel -> clear();
   else
      wordTypeLabel -> setText("<b>" + label + "</b>");
   examplesTextViewer -> setPlainText(card -> getExamples());
   commentsTextViewer -> setPlainText(card -> getComments());
   synonymsListWidget -> clear();
   synonymsListWidget -> addItems(card -> getSynonyms());
   antonymsListWidget -> clear();
   antonymsListWidget -> addItems(card -> getAntonyms());

   bool visible;
   int visibleWidgets=0;
   visible = !wordTypeLabel -> text().isEmpty();
   wordTypeCheckBox -> setVisible(visible);
   wordTypeLabel -> setVisible(visible && wordTypeCheckBox->isChecked());
   visibleWidgets += visible ? 1 : 0;

   visible = !card -> getExamples().isEmpty();
   examplesCheckBox -> setVisible(visible);
   examplesTextViewer -> setVisible(visible && examplesCheckBox->isChecked());
   visibleWidgets += visible ? 1 : 0;
   
   visible = !card -> getComments().isEmpty();
   commentsCheckBox -> setVisible(visible);
   commentsTextViewer -> setVisible(visible && commentsCheckBox->isChecked());
   visibleWidgets += visible ? 1 : 0;
   
   visible = card -> getSynonyms().size();
   synonymsCheckBox -> setVisible(visible);
   synonymsListWidget -> setVisible(visible && synonymsCheckBox->isChecked());
   visibleWidgets += visible ? 1 : 0;
   
   visible = card -> getAntonyms().size();
   antonymsCheckBox -> setVisible(visible);
   antonymsListWidget -> setVisible(visible && antonymsCheckBox->isChecked());
   visibleWidgets += visible ? 1 : 0;

   messageLabel -> setVisible(!visibleWidgets);
   if (!visibleWidgets)
      messageLabel -> setText(getNoHintsMessage());
}

void CardReviewerVocabulary::clear()
{
   textViewer -> clear();
   wordTypeLabel -> clear();
   examplesTextViewer -> clear();
   commentsTextViewer -> clear();
   synonymsListWidget -> clear();
   antonymsListWidget -> clear();
}

void CardReviewerVocabulary::play()
{
   readingID = Reader::reader() -> readHtml(getCurrentTranslationDocument() -> getText());
   CardReviewer::playingStarted();
}

void CardReviewerVocabulary::stop()
{
   Reader::reader() -> stop();
   //can cause double signal emitting
   CardReviewer::playingFinished();
}

bool CardReviewerVocabulary::isPlaying()
{
   return Reader::reader() -> isReading();
}

void CardReviewerVocabulary::on_wordTypeCheckBox_toggled(bool checked)
{
   wordTypeLabel -> setVisible(checked);
}

void CardReviewerVocabulary::on_examplesCheckBox_toggled(bool checked)
{
   examplesTextViewer -> setVisible(checked);
}

void CardReviewerVocabulary::on_commentsCheckBox_toggled(bool checked)
{
   commentsTextViewer -> setVisible(checked);
}

void CardReviewerVocabulary::on_synonymsCheckBox_toggled(bool checked)
{
   synonymsListWidget -> setVisible(checked);
   splitter -> setVisible(antonymsListWidget->isVisible() || checked);
}

void CardReviewerVocabulary::on_antonymsCheckBox_toggled(bool checked)
{
   antonymsListWidget -> setVisible(checked);
   splitter -> setVisible(checked || synonymsListWidget->isVisible());
}

void CardReviewerVocabulary::showHints()
{
   mainSplitter -> widget(1) -> show();
}

void CardReviewerVocabulary::hideHints()
{
   mainSplitter -> widget(1) -> hide();
}

void CardReviewerVocabulary::readingFinishedSlot(int id)
{
   if (readingID == id)
      CardReviewer::playingFinished();
}

void CardReviewerVocabulary::updateTextViewer()
{
   QString title = "<p style=\"color:#a0a0a0; font-size: x-small;"
                   "text-align: center; padding: 3px; background: #f0f0f0;"
                   "border-style: solid; border-width:2px; border-color:#f5f5f5;\">"
                   + (isFrontSideVisible ? tr("Front Side") : tr("Back Side")) + "</p>";
   Document document = *getCurrentTranslationDocument();
   document.setText(title + document.getText());
   textViewer -> setDocument(document);
}

void CardReviewerVocabulary::createLanguageButtons()
{
   while(languageButtons.size())
      delete languageButtons.takeAt(0);
   
   const QList<TranslationDocument*> translationDocuments = card -> getTranslationDocuments();
   for (int i=0; i<translationDocuments.size(); i++)
   {
      PrettyButton *button = new PrettyButton(Qt::Horizontal, this);
      button -> setText(translationDocuments[i]->getLanguageName());
      button -> setCheckable(true);
      connect(button, SIGNAL(hovered()), this, SLOT(languageButtonHovered()));
      connect(button, SIGNAL(clicked()), this, SLOT(languageButtonClicked()));
      languageButtons << button;
      languagesToolBar -> addWidget(button);
   }
}

void CardReviewerVocabulary::updateLanguageButtons()
{
   if (isFrontSideVisible)
      for (int i=0; i<languageButtons.size(); i++)
         languageButtons[i] -> setEnabled(i == frontSideLanguageIndex);
   else
      for (int i=0; i<languageButtons.size(); i++)
         languageButtons[i] -> setEnabled(true);
}

void CardReviewerVocabulary::selectLanguage(int index)
{
   updateLanguageButtons();

   for (int i=0; i<languageButtons.size(); i++)
      languageButtons[i] -> setChecked(false);
   languageButtons[index]->setChecked(true);
   updateTextViewer();
   if (Settings::settings() -> reviewAutoPlay)
      play();

}

void CardReviewerVocabulary::languageButtonHovered()
{
   PrettyButton *button = static_cast<PrettyButton*>(sender());
   int index = languageButtons.indexOf(button);
   if (index != -1)
   {
      if (!languageButtons[index]->isChecked())
         selectLanguage(index);
   }
}

void CardReviewerVocabulary::languageButtonClicked()
{
   PrettyButton *button = static_cast<PrettyButton*>(sender());
   button -> setChecked(true);
}

TranslationDocument* CardReviewerVocabulary::getCurrentTranslationDocument()
{
   return card -> getTranslationDocuments()[getCurrentLanguageButtonIndex()];
}

int CardReviewerVocabulary::getCurrentLanguageButtonIndex()
{
   for (int i=0; i<languageButtons.size(); i++)
      if (languageButtons[i]->isChecked())
         return i;
   qDebug("CardReviewerVocabulary::getCurrentLanguageButtonIndex() couldn't find any checked button");
   return 0;
   
}

