// +-------------------------------------------------------------------------+
// | FlashQard - An efficient way to memorize.                               |
// | Copyright (C) 2008-2009 Shahab <shahab@flashqard-project.org>           |
// +-------------------------------------------------------------------------+
// | License:                                                                |
// |                                                                         |
// | FlashQard is free software; you can redistribute it and/or modify       |
// | it under the terms of the GNU General Public License as published by    |
// | the Free Software Foundation; either version 2 of the License, or       |
// | (at your option) any later version.                                     |
// |                                                                         |
// | FlashQard is distributed in the hope that it will be useful,            |
// | but WITHOUT ANY WARRANTY; without even the implied warranty of          |
// | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the            |
// | GNU General Public License for more details.                            |
// |                                                                         |
// | You should have received a copy of the GNU General Public License       |
// | along with FlashQard; if not, write to the Free Software                |
// | Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA|
// |                                                                         |
// +-------------------------------------------------------------------------+
// | Author:  Shahab Shirazi <shahab@flashqard-project.org>                  |
// +-------------------------------------------------------------------------+

#include <QtGui>
#include "Card.h"
#include "CardRegular.h"
#include "CardVocabulary.h"
#include "CardMap.h"
#include "Settings.h"
#include "Utilities.h"

CardPosition::CardPosition(int cn)
{
   boxSection = BoxSection_Stack;
   stage=0;
   cardNumber=cn;
}

CardPosition::CardPosition(int s, int cn)
{
   boxSection = BoxSection_ActiveCards;
   stage = s;
   cardNumber = cn;
}

CardPosition::CardPosition(BoxSection bs, int s, int cn)
{
   boxSection = bs;
   stage = s;
   cardNumber = cn;
}

CardPosition::CardPosition()
{
   boxSection = BoxSection_Stack;
   stage=cardNumber=-1;
}

bool CardPosition::operator==(const CardPosition &cardPosition2) const
{
   return cardPosition2.boxSection == boxSection
      && cardPosition2.stage == stage
      && cardPosition2.cardNumber == cardNumber;
}

bool CardPosition::operator<(const CardPosition &cardPosition2) const
{
   if (boxSection == cardPosition2.boxSection && boxSection == BoxSection_Stack)
      return cardNumber<cardPosition2.cardNumber;
   else if (boxSection != cardPosition2.boxSection)
      return boxSection == BoxSection_Stack;
   else if (stage == cardPosition2.stage)
      return cardNumber<cardPosition2.cardNumber;
   else
      return stage<cardPosition2.stage;
}

//======================================================================

DocumentAttachment::DocumentAttachment(const QByteArray &ba, const QString &n, const QString &t)
{
    setByteArray(ba);
    setName(n);
    setType(t);
}

QString DocumentAttachment::getName() const
{
   return name;
}

void DocumentAttachment::setName(const QString &newName)
{
   name = newName;
}

QString DocumentAttachment::getType() const
{
   return type;
}

void DocumentAttachment::setType(const QString &_type)
{
   type = _type;
}

QByteArray DocumentAttachment::getByteArray() const
{
   return byteArray;
}

void DocumentAttachment::setByteArray(const QByteArray &newByteArray)
{
   byteArray = newByteArray;
}

//======================================================================

Document::Document()
{
   setModified(false);
}

void Document::setText(const QString &newText)
{
   text = newText;
   setModified(true);
}

QString Document::getText() const
{
   return text;
}

// QString Document::addAttachment(const QByteArray &newAttachment, const QString &type)
// {
//    QString name = Random::random() -> getARandomName();
//    addAttachment(newAttachment, name, type);
//    return name;
// }

void Document::addAttachment(const QByteArray &newAttachment, const QString &name, const QString &type)
{
   attachments.append(DocumentAttachment(newAttachment, name, type));
   setModified(true);
}

void Document::deleteAttachment(const QString &name)
{
   for (int i=0; i<attachments.size(); i++)
      if (attachments[i].getName() == name)
      {
         deleteAttachment(i);
         break;
      }
}

void Document::deleteAttachment(int index)
{
   attachments.removeAt(index);
   setModified(true);
}

void Document::deleteAllAttachments()
{
   while (attachments.size())
      deleteAttachment(0);
}

QString Document::attachmentToBase64(const QString &name) const
{
   return attachmentToBase64(findAttachment(name));
}

QString Document::attachmentToBase64(int i) const
{
   if (i<0 || i>=attachments.size())
      return QString();
   return QString(attachments[i].getByteArray().toBase64().data());
}

QImage Document::attachmentToImage(const QString &name) const
{
   return attachmentToImage(findAttachment(name));
}

QImage Document::attachmentToImage(int i) const
{
   if (i<0 || i>=attachments.size())
      return QImage();
   QImage image;
   image.loadFromData(attachments[i].getByteArray());
   return image;
}

QByteArray Document::attachmentToByteArray(const QString &name) const
{
   return attachmentToByteArray(findAttachment(name));
}

QByteArray Document::attachmentToByteArray(int i) const
{
   if (i<0 || i>=attachments.size())
      return QByteArray();
   return attachments[i].getByteArray();
}

QString Document::getAttachmentName(int i) const
{
   if (i<0 || i>=attachments.size())
      return QString();
   return attachments[i].getName();
}

QString Document::getAttachmentType(int i) const
{
   if (i<0 || i>=attachments.size())
      return QString();
   return attachments[i].getType();
}

QString Document::getAttachmentType(const QString &name) const
{
   return getAttachmentType(findAttachment(name));
}

int Document::attachmentCount() const
{
   return attachments.size();
}

bool Document::isModified() const
{
   return modified;
}

void Document::setModified(bool m)
{
   modified = m;
}

int Document::findAttachment(const QString &name) const
{
   for (int i=0; i<attachments.size(); i++)
      if (attachments[i].getName() == name)
         return i;
   return -1;
}   

//======================================================================

Answer::Answer(bool a, QDate d)
{
   answer = a;
   if (d.isNull())
      date = QDate::currentDate();
   else
      date = d;
}

bool Answer::isCorrect() const
{
   return answer;
}

QDate Answer::getDate() const
{
   return date;
}

bool Answer::operator==(const Answer &answer2) const
{
   return answer == answer2.answer && date == answer2.date;
}

//======================================================================

CardStatistics::CardStatistics(const QDate &d)
{
   if (d.isNull())
      dateCreated = QDate::currentDate();
   else
      dateCreated = d;
}

QDate CardStatistics::getDateCreated() const
{
   return dateCreated;
}

void CardStatistics::setDateCreated(const QDate &d)
{
   dateCreated = d;
}

const QList<Answer>& CardStatistics::getAnswers() const
{
   return answers;
}
      
// QList<Answer> CardStatistics::getAnswers() const
// {
//    return answers;
// }

Answer CardStatistics::getLastAnswer() const
{
   if (answers.isEmpty())
      return Answer();
   else
      return answers.last();
}

QList<Answer> CardStatistics::getTodaysAnswers() const
{
   QList<Answer> todaysAnswers;
   for (int i=0; i<answers.size(); i++)
      if (answers[i].getDate() == QDate::currentDate())
         todaysAnswers << answers[i];
   return todaysAnswers;
}

void CardStatistics::addAnswer(const Answer &newAnswer)
{
   answers.append(newAnswer);
}

void CardStatistics::clearAnswers()
{
   answers.clear();
}

int CardStatistics::correctAnswersCount() const
{
   int correctAnswers = 0;
   for (int i=0; i<answers.size(); i++)
      if (answers[i].isCorrect())
         correctAnswers++;
   return correctAnswers;
}

int CardStatistics::totalAnswersCount() const
{
   return answers.size();
}


//======================================================================

QString Card::getTitle() const
{
   QString title = makeTitle();
   if (title.isEmpty())
      title = QObject::tr("Untitled");
   return title;
}

CardStatistics& Card::getCardStatistics()
{
   return statistics;
}

const CardStatistics& Card::getCardStatistics() const
{
   return statistics;
}

void Card::setCardStatistics(const CardStatistics &newStatistics)
{
   statistics = newStatistics;
}

Card* Card::clone() const
{
   Card *newCard = cloneYourSelf();
   *newCard = *this;
   return newCard;
}

void Card::clearAnswers()
{
   statistics.clearAnswers();
}

QString Card::makeTitleFromString(const QString &string) const
{
   QString text = string.trimmed().section('\n',0,0);

   if (text.length() > 40)
      return text.mid(0,37)+QObject::tr("...");
   else
      return text;
}

//======================================================================

CardSorter::CardSorter(SortBy _sortBy, SortOrder _sortOrder)
   : sortBy(_sortBy), sortOrder(_sortOrder)
{
   
}

bool CardSorter::operator()(const Card *card1, const Card *card2)
{
   bool isLessThan;
   if (sortBy == SortBy_Title)
      isLessThan = card1->getTitle().toLower() < card2->getTitle().toLower();
   else //SortBy_CreationDate
      isLessThan = card1->getCardStatistics().getDateCreated() < card2->getCardStatistics().getDateCreated();

   return (sortOrder == SortOrder_Ascending) ? isLessThan : !isLessThan;
}

//======================================================================

CardBox::CardBox(int stages, const QString &name)
{
   boxName = name;
   for (int i=0; i<stages; i++)
      addStage();
}

CardBox::~CardBox()
{
   clear();
}

QString CardBox::getBoxName() const
{
   return boxName;
}

void CardBox::setBoxName(const QString &name)
{
   boxName = name;
}

Card* CardBox::at(const CardPosition &position)
{
   return const_cast<Card*>(getCardAt(position));
}

const Card* CardBox::at(const CardPosition &position) const
{
   return getCardAt(position);
}

const Card* CardBox::getCardAt(const CardPosition &position) const
   
{
   if (!isPositionValid(position))
      return 0;
   if (position.boxSection == BoxSection_Stack)
      return stack[position.cardNumber];
   else
      return activeCards[position.stage]->at(position.cardNumber);
}

Card* CardBox::atStack(int cardNumber)
{
   CardPosition position(cardNumber);
   return at(position);
}

const Card* CardBox::atStack(int cardNumber) const
{
   CardPosition position(cardNumber);
   return at(position);
}

Card* CardBox::atActiveCards(int stage, int cardNumber)
{
   CardPosition position(stage, cardNumber);
   return at(position);
}

const Card* CardBox::atActiveCards(int stage, int cardNumber) const
{
   CardPosition position(stage, cardNumber);
   return at(position);
}

void CardBox::addStage()
{
   activeCards.append(new CardPointerList);
}

void CardBox::removeStage(int stage)
{
   CardPosition position(stage, 0);
   while(isPositionValid(position))
      removeCard(position);
   if (stage >= 0 && stage < countStages())
      delete activeCards.takeAt(stage);
}

Card* CardBox::addCardToStack(CardType type, BeginningOrEnd position)
{
   Card *card = createCard(type);
   if (position == BeginningOrEnd_Beginning)
      stack.prepend(card);
   else
      stack.append(card);
   return card;
}

void CardBox::addCardToStack(Card *card, BeginningOrEnd position)
{
   if (position == BeginningOrEnd_Beginning)
      stack.prepend(card);
   else
      stack.append(card);
}

Card* CardBox::addCardToActiveCards(CardType type, int stage, BeginningOrEnd position)
{
   if (stage < 0 || stage >= countStages())
      return 0;
   
   Card *card = createCard(type);
   if (position == BeginningOrEnd_Beginning)
      activeCards[stage]->prepend(card);
   else
      activeCards[stage]->append(card);
   return card;
}

void CardBox::addCardToActiveCards(Card *card, int stage, BeginningOrEnd position)
{
   if (stage < 0 || stage >= countStages())
      return;
   
   if (position == BeginningOrEnd_Beginning)
      activeCards[stage]->prepend(card);
   else
      activeCards[stage]->append(card);
}

void CardBox::insertCardAt(Card *card, const CardPosition &position)
{
   if (!isPositionValid(position)
       && position.cardNumber != 0
       && !isPositionValid(CardPosition(position.boxSection, position.stage, position.cardNumber-1)))
      return;

   if (position.boxSection == BoxSection_Stack)
      stack.insert(position.cardNumber, card);
   else
      activeCards[position.stage]->insert(position.cardNumber, card);
}

Card* CardBox::createCard(CardType type)
{
   if (type == CardType_Regular)
      return new CardRegular;
   else if (type == CardType_Vocabulary)
   {
      CardVocabulary *newCardVocabulary = new CardVocabulary;
      newCardVocabulary -> getTranslationDocuments()
         << new TranslationDocument(Settings::settings()->vocabularyCardsDefaultLanguages.value(0));
      newCardVocabulary -> getTranslationDocuments()
         << new TranslationDocument(Settings::settings()->vocabularyCardsDefaultLanguages.value(1));
      if (!Settings::settings()->vocabularyCardsDefaultLanguages.value(2).isEmpty())
         newCardVocabulary -> getTranslationDocuments()
            << new TranslationDocument(Settings::settings()->vocabularyCardsDefaultLanguages.value(2));
      if (!Settings::settings()->vocabularyCardsDefaultLanguages.value(3).isEmpty())
         newCardVocabulary -> getTranslationDocuments()
            << new TranslationDocument(Settings::settings()->vocabularyCardsDefaultLanguages.value(3));
      if (!Settings::settings()->vocabularyCardsDefaultLanguages.value(4).isEmpty())
         newCardVocabulary -> getTranslationDocuments()
            << new TranslationDocument(Settings::settings()->vocabularyCardsDefaultLanguages.value(4));
         
      return newCardVocabulary;
   }
   else if (type == CardType_Map)
      return new CardMap;
   else
      return 0;
}

void CardBox::removeCard(const CardPosition &position)
{
   delete takeAt(position);
}

Card* CardBox::takeAt(const CardPosition &position)
{
   if (!isPositionValid(position))
      return 0;
   if (position.boxSection == BoxSection_Stack)
      return stack.takeAt(position.cardNumber);
   else
      return activeCards[position.stage]->takeAt(position.cardNumber);
}

CardPosition CardBox::moveCardUp(const CardPosition &position)
{
   if (!isPositionValid(position))
      return position;

   if (position.boxSection == BoxSection_Stack)
   {
      if (position.cardNumber > 0)
      {
         stack.swap(position.cardNumber, position.cardNumber-1);
         return CardPosition(position.cardNumber-1);
      }
   }
   else
   {
      if (position.cardNumber == 0)
      {
         if (position.stage>0)
         {
            moveToActiveCards(position, position.stage-1, BeginningOrEnd_End);
            return CardPosition(position.stage-1, countActiveCards(position.stage-1)-1);
         }
      }
      else
      {
         activeCards[position.stage]->swap(position.cardNumber, position.cardNumber-1);
         return CardPosition(position.stage, position.cardNumber-1);
      }
   }
   return position;
}

CardPosition CardBox::moveCardDown(const CardPosition &position)
{
   if (!isPositionValid(position))
      return position;

   if (position.boxSection == BoxSection_Stack)
   {
      if (position.cardNumber < countStack()-1)
      {
         stack.swap(position.cardNumber, position.cardNumber+1);
         return CardPosition(position.cardNumber+1);
      }
   }
   else
   {
      if (position.cardNumber == countActiveCards(position.stage)-1)
      {
         if (position.stage < countStages()-1)
         {
            moveToActiveCards(position, position.stage+1, BeginningOrEnd_Beginning);
            return CardPosition(position.stage+1, 0);
         }
      }
      else
      {
         activeCards[position.stage]->swap(position.cardNumber, position.cardNumber+1);
         return CardPosition(position.stage, position.cardNumber+1);
      }
   }
   return position;
}

void CardBox::shuffleStack()
{
   int howManyTimes = countStack()*2;
   for (int i=0; i<howManyTimes; i++)
   {
      int firstIndex = Random::random()->getARandomNumber(0, countStack()-1);
      int secondIndex = Random::random()->getARandomNumber(0, countStack()-1);
      stack.swap(firstIndex, secondIndex);
   }
}

void CardBox::shuffleActiveCards(int stage)
{
   if (stage < 0 || stage >= countStages())
      return;
   int howManyTimes = countActiveCards(stage)*2;
   for (int i=0; i<howManyTimes; i++)
   {
      int firstIndex = Random::random()->getARandomNumber(0, countActiveCards(stage)-1);
      int secondIndex = Random::random()->getARandomNumber(0, countActiveCards(stage)-1);
      activeCards[stage]->swap(firstIndex, secondIndex);
   }
}

void CardBox::clearAnswers()
{
   CardPosition pos = firstPosition();
   while (isPositionValid(pos))
   {
      at(pos) -> clearAnswers();
      pos = nextPosition(pos, MovePolicy_ThroughStackAndActiveCards);
   }
}

CardPosition CardBox::search(QString string, CardPosition position, bool isRegExp) const
{
   while(1)
   {
      position = nextPosition(position, position.boxSection == BoxSection_Stack ? MovePolicy_ThroughTheStageOrStack
                              : MovePolicy_ThroughActiveCards);
      const Card *card = at(position);
      if (card)
      {
         if (card->contains(string, isRegExp))
            return position;
      }
      else
         return position;
   }
}

void CardBox::moveTo(const CardPosition &from, const CardPosition &to)
{
   bool valid = false;
   valid = to.boxSection == BoxSection_Stack ? to.cardNumber <= countStack() && to.cardNumber >= 0
      : to.stage >= 0 && to.stage < countStages() && to.cardNumber >= 0 && to.cardNumber <= countActiveCards(to.stage);
   valid &= isPositionValid(from);
   if (!valid)
      return;
   Card *card = at(from);

   if (from.boxSection == BoxSection_Stack)
      stack.takeAt(from.cardNumber);
   else
      activeCards[from.stage]->takeAt(from.cardNumber);

   if (to.boxSection == BoxSection_Stack)
      stack.insert(to.cardNumber, card);
   else
      activeCards[to.stage]->insert(to.cardNumber, card);
}

void CardBox::moveToActiveCards(const CardPosition &from, int toStage, BeginningOrEnd insertPosition)
{
   CardPosition to(toStage, insertPosition == BeginningOrEnd_Beginning ? 0 : countActiveCards(toStage));
   moveTo(from, to);
}

void CardBox::moveToStack(const CardPosition &from, BeginningOrEnd insertPosition)
{
   CardPosition to(insertPosition == BeginningOrEnd_Beginning ? 0 : countStack());
   moveTo(from, to);
}

void CardBox::sortActiveCards(SortBy sortBy, SortOrder order, int stage)
{
   if (stage < 0 || stage >= countStages())
      return;
   
   CardSorter sorter(sortBy, order);
   qStableSort(activeCards[stage]->begin(), activeCards[stage]->end(), sorter);
}

void CardBox::sortStack(SortBy sortBy, SortOrder order)
{
   CardSorter sorter(sortBy, order);
   qStableSort(stack.begin(), stack.end(), sorter);
}

void CardBox::sortAllCards(SortBy sortBy, SortOrder order)
{
   sortStack(sortBy, order);
   for (int i=0; i<countStages(); i++)
      sortActiveCards(sortBy, order, i);
}

int CardBox::countStages() const
{
   return activeCards.size();
}

int CardBox::countActiveCards(int stage) const
{
   if (stage < -1 || stage >= activeCards.size())
      return 0;
   
   int count = 0;
   if (stage == -1)
      for (int i=0; i<activeCards.size(); i++)
         count += activeCards[i]->size();
   else
      count = activeCards[stage]->size();
   return count;
}

int CardBox::countStack() const
{
   return stack.size();
}

int CardBox::countTotalCards() const
{
   return countActiveCards() + countStack();
}

int CardBox::getTodayAnsweredCardsCount() const
{
   int reviewedCards = 0;
   QDate currentDate = QDate::currentDate();
   CardPosition position(BoxSection_ActiveCards, BeginningOrEnd_Beginning);

   while(isPositionValid(position))
   {
      if (at(position) -> getCardStatistics().getLastAnswer().getDate() == currentDate)
         reviewedCards++;
      position = nextPosition(position, MovePolicy_ThroughActiveCards);
   }
   
   return reviewedCards;
}

int CardBox::getTodayCorrectlyAnsweredCardsCount() const
{
   int correctAnswers = 0;
   QDate currentDate = QDate::currentDate();
   CardPosition position(BoxSection_ActiveCards, BeginningOrEnd_Beginning);

   while(isPositionValid(position))
   {
      Answer answer = at(position)->getCardStatistics().getLastAnswer();
      if (answer.getDate() == currentDate && answer.isCorrect())
         correctAnswers++;
      position = nextPosition(position, MovePolicy_ThroughActiveCards);
   }
   return correctAnswers;
}

int CardBox::getTodayWronglyAnsweredCardsCount() const
{
   int wrongAnswers = 0;
   QDate currentDate = QDate::currentDate();
   CardPosition position(0, 0);

   while(isPositionValid(position))
   {
      Answer answer = at(position)->getCardStatistics().getLastAnswer();
      if (answer.getDate() == currentDate && !answer.isCorrect())
         wrongAnswers++;
      position = nextPosition(position, MovePolicy_ThroughActiveCards);
   }
   return wrongAnswers;
}

void CardBox::clear()
{
   while(countStages())
      removeStage(0);

   CardPosition position(0);
   while (isPositionValid(position))
      removeCard(position);
}

CardPosition CardBox::firstPosition() const
{
   if (countStack())
      return CardPosition(0);
   else
   {
      for (int i=0; i<countStages(); i++)
         if (countActiveCards(i))
            return CardPosition(i, 0);
   }
   //return an invalid position
   return CardPosition(0,0);
}

CardPosition CardBox::nextPosition(const CardPosition &initialPosition, MovePolicy movePolicy) const
{
   int stage = initialPosition.stage;
   int cardNumber = initialPosition.cardNumber;
   BoxSection boxSection = initialPosition.boxSection;


   if (movePolicy == MovePolicy_ThroughTheStageOrStack)
   {
      cardNumber++;
   }
   
   else if (movePolicy == MovePolicy_ThroughActiveCards
            || (movePolicy == MovePolicy_ThroughStackAndActiveCards && boxSection == BoxSection_ActiveCards))
   {
      if (countActiveCards(stage) == 0 || cardNumber >= countActiveCards(stage)-1)
      {
         stage++;
         cardNumber=0;
         while (countActiveCards(stage) == 0 && stage < countStages())
            stage++;
      }
      else
         cardNumber++;
   }
   
   else //MovePolicy_ThroughStackAndActiveCards
   {
      if (boxSection == BoxSection_Stack)
      {
         cardNumber++;
         if (cardNumber >= countStack())
         {
            cardNumber = 0;
            stage = 0;
            boxSection = BoxSection_ActiveCards;
            while (countActiveCards(stage) == 0 && stage < countStages())
               stage++;
         }
      }
//       else //boxSection == BoxSection_ActiveCards
//       {
//          if (countActiveCards(stage) == 0 || cardNumber >= countActiveCards(stage)-1)
//          {
//             stage++;
//             cardNumber=0;
//             while (countActiveCards(stage) == 0 && stage < countStages())
//                stage++;
//          }
//          else
//             cardNumber++;
//       }
   }

   return CardPosition(boxSection, stage, cardNumber);
}

CardPosition CardBox::prevPosition(const CardPosition &initialPosition, MovePolicy movePolicy) const
{
   int stage = initialPosition.stage;
   int cardNumber = initialPosition.cardNumber;
   BoxSection boxSection = initialPosition.boxSection;


   if (movePolicy == MovePolicy_ThroughTheStageOrStack)
   {
      cardNumber--;
   }
   
   else if (movePolicy == MovePolicy_ThroughActiveCards)
   {
      if (countActiveCards(stage) == 0 || cardNumber <= 0)
      {
         stage--;
         while (countActiveCards(stage) == 0 && stage >= 0)
            stage--;
         cardNumber = countActiveCards(stage)-1;
      }
      else
         cardNumber--;
   }
   
   else //MovePolicy_ThroughStackAndActiveCards
   {
      if (boxSection == BoxSection_Stack)
      {
         cardNumber--;
      }
      else //boxSection == BoxSection_ActiveCards
      {
         if (countActiveCards(stage) == 0 || cardNumber <= 0)
         {
            stage--;
            while (countActiveCards(stage) == 0 && stage >= 0)
               stage--;
            cardNumber = countActiveCards(stage)-1;
         }
         else
            cardNumber--;

         if (stage == -1)
         {
            cardNumber = countStack()-1;
            boxSection = BoxSection_Stack;
         }
      }
   }

   return CardPosition(boxSection, stage, cardNumber);
}

bool CardBox::isPositionValid(const CardPosition &position) const
{
   if (position.boxSection == BoxSection_Stack)
      return position.cardNumber >= 0 && position.cardNumber < countStack();
   else
      return position.stage >= 0 && position.stage < countStages()
      && position.cardNumber >= 0 && position.cardNumber < countActiveCards(position.stage);
}

CardPosition CardBox::positionOfCard(const Card *card) const
{
   for (int i=0; i<activeCards.size(); i++)
      for (int j=0; j<activeCards[i]->size(); j++)
         if (card == activeCards[i]->at(j))
            return CardPosition(i, j);
   for (int i=0; i<stack.size(); i++)
      if (card == stack[i])
         return CardPosition(i);
   return CardPosition();
}

//======================================================================

void CardBoxesInformation::setAuthor(const QString &a)
{
   author = a;
}

QString CardBoxesInformation::getAuthor() const
{
   return author;
}

void CardBoxesInformation::setAuthorEmail(const QString &a)
{
   authorEmail = a;
}

QString CardBoxesInformation::getAuthorEmail() const
{
   return authorEmail;
}

void CardBoxesInformation::setLicense(const QString &l)
{
   license = l;
}

QString CardBoxesInformation::getLicense() const
{
   return license;
}

void CardBoxesInformation::setComment(const QString &c)
{
   comment = c;
}

QString CardBoxesInformation::getComment() const
{
   return comment;
}

//======================================================================

CardBoxes::CardBoxes()
{

}

CardBoxes::CardBoxes(const CardBoxes &cardBoxes)
{
   //use = operator
   *this = cardBoxes;
}

CardBoxes::~CardBoxes()
{
   clear();
}

void CardBoxes::clear()
{
   while (!isEmpty())
      delete takeFirst();
   information = CardBoxesInformation();
}

const CardBoxes& CardBoxes::operator=(const CardBoxes &cardBoxes)
{
   clear();
   for (int i=0; i<cardBoxes.size(); i++)
   {
      append(new CardBox(0, cardBoxes[i]->getBoxName()));
      for (int j=0; j<cardBoxes[i]->countStages(); j++)
      {
         at(i) -> addStage();
         for (int k=0; k<cardBoxes[i]->countActiveCards(j); k++)
         {
            Card *card = cardBoxes[i]->atActiveCards(j, k) -> clone();
            at(i) -> addCardToActiveCards(card, j, BeginningOrEnd_End);
         }
      }

      for (int z=0; z<cardBoxes[i]->countStack(); z++)
      {
         Card *card = cardBoxes[i]->atStack(z) -> clone();
         at(i) -> addCardToStack(card, BeginningOrEnd_End);
      }
   }
   setInformation(cardBoxes.getInformation());
   return *this;
}

void CardBoxes::setInformation(const CardBoxesInformation &_information)
{
   information = _information;
}

const CardBoxesInformation& CardBoxes::getInformation() const
{
   return information;
}

