// +-------------------------------------------------------------------------+
// | FlashQard - An efficient way to memorize.                               |
// | Copyright (C) 2008-2009 Shahab <shahab@flashqard-project.org>           |
// +-------------------------------------------------------------------------+
// | License:                                                                |
// |                                                                         |
// | FlashQard is free software; you can redistribute it and/or modify       |
// | it under the terms of the GNU General Public License as published by    |
// | the Free Software Foundation; either version 2 of the License, or       |
// | (at your option) any later version.                                     |
// |                                                                         |
// | FlashQard is distributed in the hope that it will be useful,            |
// | but WITHOUT ANY WARRANTY; without even the implied warranty of          |
// | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the            |
// | GNU General Public License for more details.                            |
// |                                                                         |
// | You should have received a copy of the GNU General Public License       |
// | along with FlashQard; if not, write to the Free Software                |
// | Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA|
// |                                                                         |
// +-------------------------------------------------------------------------+
// | Author:  Shahab Shirazi <shahab@flashqard-project.org>                  |
// +-------------------------------------------------------------------------+

#ifndef CARD_H
#define CARD_H

#include <QString>
#include <QList>
#include <QTextDocument>
#include <QTextCursor>
#include <QDate>

class CardBox;

///QList's at(int) function only returns a const reference
///returning an editable reference can be useful. so:
template <typename T>
class MyQList : public QList<T>
{
   public:
      MyQList(){}
      MyQList(const QList<T> &other) { *this = other; }
      ~MyQList(){}
      T& at(int i) {return (*this)[i];}
      const T& at(int i) const {return (*this)[i];}
};

//======================================================================

enum BoxSection{BoxSection_ActiveCards, BoxSection_Stack};
enum MovePolicy{MovePolicy_ThroughTheStageOrStack,
                MovePolicy_ThroughActiveCards,
                MovePolicy_ThroughStackAndActiveCards};

class CardPosition
{
   public:
      CardPosition(int);
      CardPosition(int, int);
      CardPosition(BoxSection, int, int);
      CardPosition();

      bool operator==(const CardPosition&) const;
      bool operator<(const CardPosition&) const;

      BoxSection boxSection;
      int stage;
      int cardNumber;
};

//======================================================================

#define ATTACHMENT_IMAGE "image"

class DocumentAttachment
{
   public:
      DocumentAttachment(const QByteArray&, const QString&, const QString&);
      QString getName() const;
      void setName(const QString&);
      QString getType() const;
      void setType(const QString&);
      QByteArray getByteArray() const;
      void setByteArray(const QByteArray&);

   private:
      QByteArray byteArray;
      QString name;
      QString type;
};

//======================================================================

class Document
{
   public:
      Document();
      virtual ~Document(){}
      void setText(const QString&);
      QString getText() const;

//      QString addAttachment(const QByteArray&, const QString&);
      void addAttachment(const QByteArray&, const QString&, const QString&);
      void deleteAttachment(const QString&);
      void deleteAttachment(int);
      void deleteAllAttachments();
      QString attachmentToBase64(const QString&) const;
      QString attachmentToBase64(int) const;
      QImage attachmentToImage(const QString&) const;
      QImage attachmentToImage(int) const;
      QByteArray attachmentToByteArray(const QString&) const;
      QByteArray attachmentToByteArray(int) const;
      QString getAttachmentName(int) const;
      QString getAttachmentType(int) const;
      QString getAttachmentType(const QString&) const;

      int attachmentCount() const;

      bool isModified() const;
      void setModified(bool);

   private:
      QString text;
      QList<DocumentAttachment> attachments;
      bool modified;

      int findAttachment(const QString&) const;
};

//======================================================================

class Answer
{
   public:
      Answer(){}
      Answer(bool, QDate = QDate());
      bool isCorrect() const;
      QDate getDate() const;

      bool operator==(const Answer&) const;

   private:
      bool answer;
      QDate date;
};

//======================================================================

class CardStatistics
{
   public:
      CardStatistics(const QDate& = QDate());
      QDate getDateCreated() const;
      void setDateCreated(const QDate&);
      
//      QList<Answer> getAnswers() const;
      const QList<Answer>& getAnswers() const;
      Answer getLastAnswer() const;
      QList<Answer> getTodaysAnswers() const;
      void addAnswer(const Answer&);
      void clearAnswers();

      int correctAnswersCount() const;
      int totalAnswersCount() const;

   private:
      QDate dateCreated;
      QList<Answer> answers;
};

//======================================================================

enum CardType {CardType_Regular, CardType_Vocabulary, CardType_Map};

class Card
{
   public:
      Card() {}
      virtual ~Card() {}

      QString getTitle() const;
      virtual CardType getType() const = 0;
      virtual bool contains(const QString&, bool) const = 0;
      
      CardStatistics& getCardStatistics();
      const CardStatistics& getCardStatistics() const;
      void setCardStatistics(const CardStatistics&);
      void clearAnswers();

      Card* clone() const;

   protected:
      virtual Card* cloneYourSelf() const = 0;
      virtual QString makeTitle() const = 0;
      QString makeTitleFromString(const QString&) const;
      
   private:
      CardStatistics statistics;
};

//======================================================================

enum SortBy { SortBy_Title, SortBy_CreationDate };
enum SortOrder {SortOrder_Ascending, SortOrder_Descending};

class CardSorter
{
   public:
      CardSorter(SortBy, SortOrder);
      bool operator()(const Card*, const Card*);

   private:
      SortBy sortBy;
      SortOrder sortOrder;
};

typedef MyQList<Card*> CardPointerList;
enum BeginningOrEnd {BeginningOrEnd_Beginning, BeginningOrEnd_End};

class CardBox
{
   public:
      CardBox(int, const QString& = QObject::tr("Untitled"));
      ~CardBox();
      void setBoxName(const QString&);
      QString getBoxName() const;

      Card* at(const CardPosition&);
      const Card* at(const CardPosition&) const;

      Card* atStack(int);
      const Card* atStack(int) const;

      Card* atActiveCards(int, int);
      const Card* atActiveCards(int, int) const;
      
      void addStage();
      void removeStage(int);

      Card* addCardToStack(CardType, BeginningOrEnd = BeginningOrEnd_Beginning);
      void addCardToStack(Card*, BeginningOrEnd = BeginningOrEnd_Beginning);
      Card* addCardToActiveCards(CardType, int=0, BeginningOrEnd = BeginningOrEnd_Beginning);
      void addCardToActiveCards(Card*, int=0, BeginningOrEnd = BeginningOrEnd_Beginning);
      void insertCardAt(Card*, const CardPosition&);
      
      void removeCard(const CardPosition&);
      Card* takeAt(const CardPosition&);
      CardPosition moveCardUp(const CardPosition&);
      CardPosition moveCardDown(const CardPosition&);
      void shuffleStack();
      void shuffleActiveCards(int);
      void clearAnswers();

      //Searches only in activeCards if CardPosition's boxSection == BoxSection_ActiveCards
      //or, searches only in stack if CardPosition's boxSection == BoxSection_Stack
      CardPosition search(QString, CardPosition, bool) const;
      void moveTo(const CardPosition&, const CardPosition&);
      void moveToActiveCards(const CardPosition&, int, BeginningOrEnd);
      void moveToStack(const CardPosition&, BeginningOrEnd);

      void sortActiveCards(SortBy, SortOrder, int);
      void sortStack(SortBy, SortOrder);
      void sortAllCards(SortBy, SortOrder);

      int countStages() const;
      //-1 means all active cards
      int countActiveCards(int=-1) const;
      int countStack() const;
      int countTotalCards() const;
      
      int getTodayAnsweredCardsCount() const;
      int getTodayCorrectlyAnsweredCardsCount() const;
      int getTodayWronglyAnsweredCardsCount() const;
      void clear();

      CardPosition firstPosition() const;
      CardPosition nextPosition(const CardPosition&, MovePolicy) const;
      CardPosition prevPosition(const CardPosition&, MovePolicy) const;
      bool isPositionValid(const CardPosition&) const;
      CardPosition positionOfCard(const Card*) const;
      
   private:
      QString boxName;
      MyQList<CardPointerList*> activeCards;
      CardPointerList stack;

      Card* createCard(CardType);
      const Card* getCardAt(const CardPosition&) const;
};

//======================================================================

class CardBoxesInformation
{
   public:
      void setAuthor(const QString&);
      QString getAuthor() const;
      void setAuthorEmail(const QString&);
      QString getAuthorEmail() const;
      void setLicense(const QString&);
      QString getLicense() const;
      void setComment(const QString&);
      QString getComment() const;

   private:
      QString author;
      QString authorEmail;
      QString license;
      QString comment;
};

class CardBoxes : public MyQList<CardBox*>
{
   public:
      CardBoxes();
      CardBoxes(const CardBoxes&);
      ~CardBoxes();
      const CardBoxes& operator=(const CardBoxes&);
      void clear();

      void setInformation(const CardBoxesInformation&);
      const CardBoxesInformation& getInformation() const;

   private:
      CardBoxesInformation information;
};


#endif

