// +-------------------------------------------------------------------------+
// | FlashQard - An efficient way to memorize.                               |
// | Copyright (C) 2008-2009 Shahab <shahab@flashqard-project.org>           |
// +-------------------------------------------------------------------------+
// | License:                                                                |
// |                                                                         |
// | FlashQard is free software; you can redistribute it and/or modify       |
// | it under the terms of the GNU General Public License as published by    |
// | the Free Software Foundation; either version 2 of the License, or       |
// | (at your option) any later version.                                     |
// |                                                                         |
// | FlashQard is distributed in the hope that it will be useful,            |
// | but WITHOUT ANY WARRANTY; without even the implied warranty of          |
// | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the            |
// | GNU General Public License for more details.                            |
// |                                                                         |
// | YOU should have received a copy of the GNU General Public License       |
// | along with FlashQard; if not, write to the Free Software                |
// | Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA|
// |                                                                         |
// +-------------------------------------------------------------------------+
// | Author:  Shahab Shirazi <shahab@flashqard-project.org>                  |
// +-------------------------------------------------------------------------+

#include "CardVocabulary.h"
#include "Utilities.h"


TranslationDocument::TranslationDocument(const QString &newLanguage)
{
   languageName = newLanguage;
   setModified(true);
}

void TranslationDocument::setLanguageName(const QString &newLanguage)
{
   languageName = newLanguage;
   setModified(true);
}

QString TranslationDocument::getLanguageName() const
{
   return languageName;
}

//----------------------------------------------------------------------

CardVocabulary::CardVocabulary()
{
   wordType = 0;
}

CardVocabulary::~CardVocabulary()
{
   while (translationDocuments.size())
      delete translationDocuments.takeAt(0);
}

CardType CardVocabulary::getType() const
{
   return CardType_Vocabulary;
}

QString CardVocabulary::makeTitle() const
{
   if (translationDocuments[0]->isModified())
   {
      title = makeTitleFromString(HtmlParser::htmlParser()->htmlToPlainText(translationDocuments[0]->getText()));
      translationDocuments[0]->setModified(false);
   }
   return title;
}

bool CardVocabulary::contains(const QString &string, bool isRegExp) const
{
   for (int i=0; i<translationDocuments.size(); i++)
      if (HtmlParser::htmlParser()->contains(translationDocuments[i]->getText(), string, isRegExp))
         return true;

   QRegExp regExp = isRegExp ? QRegExp(string) : QRegExp(string, Qt::CaseInsensitive, QRegExp::Wildcard);

   return (examples.contains(regExp)
           || comments.contains(regExp)
           || synonyms.join("\n").contains(regExp)
           || antonyms.join("\n").contains(regExp)
           || userDefinedWordType.contains(regExp));
}

QList<TranslationDocument*>& CardVocabulary::getTranslationDocuments()
{
   return translationDocuments;
}

const QList<TranslationDocument*>& CardVocabulary::getTranslationDocuments() const
{
   return translationDocuments;
}

int CardVocabulary::getWordType() const
{
   return wordType;
}

QString CardVocabulary::wordTypeToString(const QString &separator) const
{
   QStringList strings;
   if (wordType & WordType_Noun)
      strings << QObject::tr("Noun");
   if (wordType & WordType_Verb)
      strings << QObject::tr("Verb");
   if (wordType & WordType_Adjective)
      strings << QObject::tr("Adjective");
   if (wordType & WordType_Adverb)
      strings << QObject::tr("Adverb");
   if (wordType & WordType_Other)
      strings << userDefinedWordType;
   return strings.join(separator);
}

void CardVocabulary::setWordType(int wt)
{
   wordType = wt;
}

int CardVocabulary::wordTypeFromString(const QString &string, const QString &separator)
{
   int wordType = 0;
   QStringList list = string.split(QRegExp("\\s*,\\s*"));

   for (int i=0; i<list.size(); i++)
   {
      if (list[i].compare("Noun", Qt::CaseInsensitive) == 0)
         wordType |= WordType_Noun;
      else if (list[i].compare("Verb", Qt::CaseInsensitive) == 0)
         wordType |= WordType_Verb;
      else if (list[i].compare("Adjective", Qt::CaseInsensitive) == 0)
         wordType |= WordType_Adjective;
      else if (list[i].compare("Adverb", Qt::CaseInsensitive) == 0)
         wordType |= WordType_Adverb;
   }
   return wordType;
}

QString CardVocabulary::getUserDefinedWordType() const
{
   return userDefinedWordType;
}

void CardVocabulary::setUserDefinedWordType(const QString &wt)
{
   userDefinedWordType = wt;
}

QStringList CardVocabulary::getSynonyms() const
{
   return synonyms;
}

void CardVocabulary::setSynonyms(const QStringList &newSynonyms)
{
   synonyms = newSynonyms;
}

QStringList CardVocabulary::getAntonyms() const
{
   return antonyms;
}

void CardVocabulary::setAntonyms(const QStringList &newAntonyms)
{
   antonyms = newAntonyms;
}

QString CardVocabulary::getExamples() const
{
   return examples;
}

void CardVocabulary::setExamples(const QString &newExamples)
{
   examples = newExamples;
}

QString CardVocabulary::getComments() const
{
   return comments;
}

void CardVocabulary::setComments(const QString &newComments)
{
   comments = newComments;
}

Card* CardVocabulary::cloneYourSelf() const
{
   CardVocabulary *newCard = new CardVocabulary(*this);
   newCard -> translationDocuments.clear();
   for (int i=0; i<translationDocuments.size(); i++)
      newCard -> translationDocuments << new TranslationDocument(*translationDocuments[i]);
   return newCard;
}





