// +-------------------------------------------------------------------------+
// | FlashQard - An efficient way to memorize.                               |
// | Copyright (C) 2008-2009 Shahab <shahab@flashqard-project.org>           |
// +-------------------------------------------------------------------------+
// | License:                                                                |
// |                                                                         |
// | FlashQard is free software; you can redistribute it and/or modify       |
// | it under the terms of the GNU General Public License as published by    |
// | the Free Software Foundation; either version 2 of the License, or       |
// | (at your option) any later version.                                     |
// |                                                                         |
// | FlashQard is distributed in the hope that it will be useful,            |
// | but WITHOUT ANY WARRANTY; without even the implied warranty of          |
// | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the            |
// | GNU General Public License for more details.                            |
// |                                                                         |
// | You should have received a copy of the GNU General Public License       |
// | along with FlashQard; if not, write to the Free Software                |
// | Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA|
// |                                                                         |
// +-------------------------------------------------------------------------+
// | Author:  Shahab Shirazi <shahab@flashqard-project.org>                  |
// +-------------------------------------------------------------------------+

#include "FileReader.h"
#include "CardRegular.h"
#include "CardVocabulary.h"
#include "CardMap.h"
#include "FQ2ImportDialog.h"
#include "CSVImportDialog.h"
#include "Utilities.h"
#include <QString>
#include <QFile>
#include <QMessageBox>
#include <QtXml>

FileReaderSettings::FileReaderSettings()
{
   categoryMode = Category_Automatic;
   stageStackMode = StageStack_Automatic;
   existingCategoryNumber = 0;
   stageNumber = 0;
   importStatistics = true;
   clearExistingCards = true;
}

//----------------------------------------------------------------------

bool FileReader::readFromFile(const QString &fileName, CardBoxes *cardBoxes, const FileReaderSettings &config)
{
   QFile file(fileName);
   //NOTE: File is opened in Text mode
   if (!file.open(QIODevice::ReadOnly))
   {
      QMessageBox::warning(0, QObject::tr("Open file"),
                           QObject::tr("Can not open this file:\n%1")
                           .arg(fileName));
      return false;
   }

   CardBoxes newCardBoxes;
   bool success = read(&file, &newCardBoxes);
   if (!success)
      return false;

   if (config.clearExistingCards)
      cardBoxes -> clear();

   bool automaticCategory = config.categoryMode == FileReaderSettings::Category_Automatic;
   int categoryNumber;
   switch(config.categoryMode)
   {
      case FileReaderSettings::Category_Create:
         cardBoxes -> append(new CardBox(5, config.newCategoryName));
         categoryNumber = cardBoxes -> size()-1;
         break;

      case FileReaderSettings::Category_AddTo:
         categoryNumber = config.existingCategoryNumber;
         if (categoryNumber == -1)
         {
            cardBoxes -> append(new CardBox(5));
            categoryNumber = 0;
         }
         break;

      default:
         categoryNumber = 0;
   }
   FileReaderSettings::StageStack stageStackMode = config.stageStackMode;
   int stageNumber = config.stageNumber;
   bool importStatistics = config.importStatistics;

   for (int i=0; i<newCardBoxes.size(); i++)
   {
      int category = categoryNumber;
      if (automaticCategory)
      {
         cardBoxes -> append(new CardBox(newCardBoxes[i]->countStages(), newCardBoxes[i]->getBoxName()));
         category = cardBoxes -> size()-1;
      }
         
      CardPosition pos = newCardBoxes[i]->firstPosition();
      while(newCardBoxes[i]->isPositionValid(pos))
      {
         int stage = stageNumber;
         if (stageStackMode == FileReaderSettings::StageStack_Automatic)
            stage = pos.stage;
            
         while(stage >= cardBoxes -> at(category) -> countStages())
            cardBoxes -> at(category) -> addStage();

         if (stageStackMode == FileReaderSettings::StageStack_AddToStack
             || (stageStackMode == FileReaderSettings::StageStack_Automatic && pos.boxSection == BoxSection_Stack))
         {
            Card *card = cloneCard(newCardBoxes[i]->at(pos), importStatistics);
            cardBoxes -> at(category) -> addCardToStack(card, BeginningOrEnd_End);
         }
         else
         {
            Card *card = cloneCard(newCardBoxes[i]->at(pos), importStatistics);
            cardBoxes -> at(category) -> addCardToActiveCards(card, stage, BeginningOrEnd_End);
         }

         pos = newCardBoxes[i] -> nextPosition(pos, MovePolicy_ThroughStackAndActiveCards);
      }
   }
   cardBoxes -> setInformation(newCardBoxes.getInformation());
   return true;
}

Card* FileReader::cloneCard(const Card *card, bool importStatistics) const
{
   Card *newCard = card -> clone();
   if (!importStatistics)
      newCard -> clearAnswers();
   return newCard;
}

bool FileReader::read(QFile *file, CardBoxes *cardBoxes)
{
   if (isItFQ(file))
      return readFQ(file, cardBoxes);
   else if (isItKVTML(file))
      return readKVTML(file, cardBoxes);
   
   else if (readCSV(file, cardBoxes))
      return true;
   else
   {
      QMessageBox::warning(0, QObject::tr("Reading Failed"),
                           QObject::tr("Could not read the file."));
      return false;
   }
}

bool FileReader::readFQ(QFile *file, CardBoxes *cardBoxes)
{
   QString errorStr;
   int errorLine;
   int errorColumn;
   QDomDocument doc;

   if (!doc.setContent(file, true, &errorStr, &errorLine, &errorColumn))
   {
      QMessageBox::warning(0, QObject::tr("Parser"),
                           QObject::tr("Parse error at line %1, column %2:\n%3")
                           .arg(errorLine)
                           .arg(errorColumn)
                           .arg(errorStr));
      return false;
   }
   QDomElement root = doc.documentElement();
   if (root.tagName() != "flashcards")
   {
      QMessageBox::warning(0, QObject::tr("Parser"),
                           QObject::tr("This is not a FlashQard file."));
      return false;
   }
   
   QString writerVersion = root.attribute("writerversion");
   QStringList flashqardVersion_separated = root.attribute("flashqardversion").split(".", QString::SkipEmptyParts);
   bool readingFromOldFiles = false;
   if (!flashqardVersion_separated.size()
       || (flashqardVersion_separated.value(0).toInt() == 0 && flashqardVersion_separated.value(1).toInt() <= 12))
   {
      readingFromOldFiles = true;
      qDebug("flashqardversion is less than 0.12.x");
   }

   if (writerVersion == "1" || writerVersion.isEmpty())
      return ParserFQ_1().read(&doc, cardBoxes);
   else if (writerVersion == "2")
   {
      ParserFQ_2 parseFQ_2;
      parseFQ_2.setReadingFromOldFiles(readingFromOldFiles);
      return parseFQ_2.read(&doc, cardBoxes);
   }
   else if (writerVersion == "3")
   {
      ParserFQ_3 parseFQ_3;
      parseFQ_3.setReadingFromOldFiles(readingFromOldFiles);
      return parseFQ_3.read(&doc, cardBoxes);
   }
   else
   {
      QMessageBox::warning(0, QObject::tr("Parser"),
                           QObject::tr("This file has been written by a newer version of this software."
                                       " It is not supported in this version."));
      return false;
   }
}

bool FileReader::readCSV(QFile *file, CardBoxes *cardBoxes)
{
   QTextStream stream(file);
   return ParserCSV().read(&stream, cardBoxes);
}

bool FileReader::readKVTML(QFile *file, CardBoxes *cardBoxes)
{
   QString errorStr;
   int errorLine;
   int errorColumn;
   QDomDocument doc;

   if (!doc.setContent(file, true, &errorStr, &errorLine, &errorColumn))
   {
      QMessageBox::warning(0, QObject::tr("Parser"),
                           QObject::tr("Parse error at line %1, column %2:\n%3")
                           .arg(errorLine)
                           .arg(errorColumn)
                           .arg(errorStr));
      return false;
   }

   QDomElement root = doc.documentElement();
   if (root.tagName() != "kvtml")
   {
      QMessageBox::warning(0, QObject::tr("Parser"),
                           QObject::tr("This is not a kvtml file."));
      return false;
   }
   QString version = root.attribute("version");
   if (version == "2.0")
      return ParserKVTML_2_0().read(&doc, cardBoxes);
   else
   {
      QMessageBox::warning(0, QObject::tr("Parser"),
                           QObject::tr("This file format is not supported.\n"
                                       "Only version 2 of KVTML file format is supported.\n"));
      return false;
   }
}

bool FileReader::isItFQ(QFile *file) const
{
   QDomDocument doc;
   if (!doc.setContent(file))
   {
      file->seek(0);
      return false;
   }
   QDomElement root = doc.documentElement();
   bool result = root.tagName() == "flashcards";
   file->seek(0);
   return result;
}

bool FileReader::isItKVTML(QFile *file) const
{
   QDomDocument doc;
   if (!doc.setContent(file))
   {
      file->seek(0);
      return false;
   }
   QDomElement root = doc.documentElement();
   bool result = root.tagName() == "kvtml";
   file->seek(0);
   return result;
}

//======================================================================

bool ParserFQ_1::read(QDomDocument *doc, CardBoxes *cardBoxes)
{
   QDomElement root = doc -> documentElement();
   QDomNode node = root.firstChild();
   while (!node.isNull())
   {
      if (node.toElement().tagName() == "box")
      {
         CardBox *box = new CardBox(0);
         parseBox(node.toElement(), box);
         cardBoxes->append(box);
      }
      node = node.nextSibling();
   }
   return true;
}

void ParserFQ_1::parseBox(const QDomElement &element, CardBox *box)
{
   box->setBoxName(element.attribute("name"));
   QDomNode node = element.firstChild();
   for (int i=0; !node.isNull(); i++)
   {
      box->addStage();
      if (node.toElement().tagName() == "stage")
         parseStage(node.toElement(), box, i);
      node = node.nextSibling();
   }
}

void ParserFQ_1::parseStage(const QDomElement &element, CardBox *box, int stage)
{
   QDomNode node = element.firstChild();
   while (!node.isNull())
   {
      if (node.toElement().tagName() == "card")
      {
         CardRegular *card = new CardRegular;
         parseCard(node.toElement(), card);
         box->addCardToActiveCards(card, stage, BeginningOrEnd_End);
      }
      node = node.nextSibling();
   }
}

void ParserFQ_1::parseCard(const QDomElement &element, CardRegular *card)
{
   QDomNode node = element.firstChild();
   while (!node.isNull())
   {
      if (node.toElement().tagName() == "frontsidetext")
      {
         QDomNode childNode = node.firstChild();
         if (!childNode.isNull())
            if (childNode.nodeType() == QDomNode::TextNode)
               card->getFrontSideDocument()->setText(childNode.toText().data());
      }
      else if (node.toElement().tagName() == "backsidetext")
      {
         QDomNode childNode = node.firstChild();
         if (!childNode.isNull())
            if (childNode.nodeType() == QDomNode::TextNode)
               card->getBackSideDocument()->setText(childNode.toText().data());
      }
      else if (node.toElement().tagName() == "frontsideattachment")
      {
         QDomNode childNode = node.firstChild();
         if (!childNode.isNull())
            if(childNode.nodeType() == QDomNode::TextNode)
            {
               card->getFrontSideDocument()->addAttachment(QByteArray::fromBase64(childNode.toText().data().toAscii().constData()),
                                                           node.toElement().attribute("name"), node.toElement().attribute("type"));
            }
      }
      else if (node.toElement().tagName() == "backsideattachment")
      {
         QDomNode childNode = node.firstChild();
         if (!childNode.isNull())
            if(childNode.nodeType() == QDomNode::TextNode)
            {
               card->getBackSideDocument()->addAttachment(QByteArray::fromBase64(childNode.toText().data().toAscii().constData()),
                                                          node.toElement().attribute("name"), node.toElement().attribute("type"));
            }
      }
      else if (node.toElement().tagName() == "dateCreated")
      {
         QDomNode childNode = node.firstChild();
         if (!childNode.isNull())
            if (childNode.nodeType() == QDomNode::TextNode)
               card->getCardStatistics().setDateCreated(QDate::fromString(childNode.toText().data(), "d.M.yyyy"));
      }
      else if (node.toElement().tagName() == "answer")
      {
         QDomNode childNode = node.firstChild();
         if (!childNode.isNull())
            if (childNode.nodeType() == QDomNode::TextNode)
            {
               card->getCardStatistics().addAnswer(Answer(childNode.toText().data() == "true" ? true : false,
                                                         QDate::fromString(node.toElement().attribute("date"), "d.M.yyyy")));
            }
      }
      
      node = node.nextSibling();
   }
}


//======================================================================

ParserFQ_2::ParserFQ_2()
{
   readingFromOldFiles = false;
}

bool ParserFQ_2::read(QDomDocument *doc, CardBoxes *cardBoxes)
{
   FQ2ImportDialog dialog;
   if (!dialog.exec())
      return false;
   wordLanguage = dialog.getWordLanguage();
   meaningLanguage = dialog.getMeaningLanguage();
   return ParserFQ_1::read(doc, cardBoxes);
}

void ParserFQ_2::setReadingFromOldFiles(bool old)
{
   readingFromOldFiles = old;
}

void ParserFQ_2::parseStage(const QDomElement &element, CardBox *box, int stage)
{
   QDomNode node = element.firstChild();
   while (!node.isNull())
   {
      if (node.toElement().tagName() == "card")
      {
         Card *card = 0;
         QString cardType = node.toElement().attribute("type");
         if (cardType == "regular")
            card = parseCardRegular(node.toElement());
         else if (cardType == "vocabulary")
            card = parseCardVocabulary(node.toElement());

         if (card)
            box->addCardToActiveCards(card, stage, BeginningOrEnd_End);
      }
      node = node.nextSibling();
   }
}

Card* ParserFQ_2::parseCardRegular(const QDomElement &element)
{
   CardRegular *card = new CardRegular;
   QDomNode node = element.firstChild();
   while (!node.isNull())
   {
      QString tagName = node.toElement().tagName();
      if (tagName == "frontsidedocument")
         parseDocument(node.toElement(), card->getFrontSideDocument());
      else if (tagName == "backsidedocument")
         parseDocument(node.toElement(), card->getBackSideDocument());
      else if (tagName == "statistics")
         parseCardStatistics(node.toElement(), card->getCardStatistics());
      else if (tagName == "comments")
         card -> setComments(getText(node.firstChild()));

      node = node.nextSibling();
   }
   return card;
}

Card* ParserFQ_2::parseCardVocabulary(const QDomElement &element)
{
   CardVocabulary *card = new CardVocabulary;
   card -> getTranslationDocuments() << new TranslationDocument(wordLanguage);
   card -> getTranslationDocuments() << new TranslationDocument(meaningLanguage);
   
   QDomNode node = element.firstChild();
   while (!node.isNull())
   {
      QString tagName = node.toElement().tagName();
      if (tagName == "worddocument")
         parseDocument(node.toElement(), card->getTranslationDocuments()[0]);
      else if (tagName == "meaningdocument")
         parseDocument(node.toElement(), card->getTranslationDocuments()[1]);
      else if (tagName == "statistics")
         parseCardStatistics(node.toElement(), card->getCardStatistics());
      else if (tagName == "examples")
         card -> setExamples(getText(node.firstChild()));
      else if (tagName == "wordtype")
      {
         int wordType = 0;
         QStringList list = getText(node.firstChild()).split(QRegExp("\\s*,\\s*"));
         for (int i=0; i<list.size(); i++)
         {
            if (list[i].compare("Noun", Qt::CaseInsensitive) == 0)
               wordType |= WordType_Noun;
            else if (list[i].compare("Verb", Qt::CaseInsensitive) == 0)
               wordType |= WordType_Verb;
            else if (list[i].compare("Adjective", Qt::CaseInsensitive) == 0)
               wordType |= WordType_Adjective;
            else if (list[i].compare("Adverb", Qt::CaseInsensitive) == 0)
               wordType |= WordType_Adverb;
            else if (list[i].compare("Other", Qt::CaseInsensitive) == 0)
               wordType |= WordType_Other;
         }
         card -> setWordType(wordType);
      }
      else if (tagName == "userdefinedwordtype")
         card -> setUserDefinedWordType(getText(node.firstChild()));
      else if (tagName == "comments")
         card -> setComments(getText(node.firstChild()));
      else if (tagName == "synonyms")
         card -> setSynonyms(getText(node.firstChild()).split(QRegExp("\\s*,\\s*"), QString::SkipEmptyParts));
      else if (tagName == "antonyms")
         card -> setAntonyms(getText(node.firstChild()).split(QRegExp("\\s*,\\s*"), QString::SkipEmptyParts));

      node = node.nextSibling();
   }
   return card;
}

void ParserFQ_2::parseDocument(const QDomElement &element, Document *document)
{
   QMap<QString, QString> newNames; //for readingFromOldFiles
   
   QDomNode node = element.firstChild();
   while (!node.isNull())
   {
      QString tagName = node.toElement().tagName();
      if (tagName == "html")
      {
         QString html = getText(node.firstChild());
         if (readingFromOldFiles)
         {
            newNames = substitudeNewNamesForImages(html);
            html = removeStyle(html);
         }
         document->setText(html);
      }
      else if (tagName == "attachment")
      {
         QString name = node.toElement().attribute("name");
         QString type = node.toElement().attribute("type");
         if (readingFromOldFiles)
         {
            name = newNames[name];
            type = ATTACHMENT_IMAGE;
         }
         document->addAttachment(QByteArray::fromBase64(getText(node.firstChild()).toAscii().constData()), name, type);
      }

      node = node.nextSibling();
   }
}

void ParserFQ_2::parseCardStatistics(const QDomElement &element, CardStatistics &statistics)
{
   QDomNode node = element.firstChild();
   while (!node.isNull())
   {
      QString tagName = node.toElement().tagName();
      if (tagName == "dateCreated")
         statistics.setDateCreated(QDate::fromString(getText(node.firstChild()), "d.M.yyyy"));
      else if (tagName == "answer")
         statistics.addAnswer(Answer(getText(node.firstChild()) == "true" ? true : false,
                                     QDate::fromString(node.toElement().attribute("date"), "d.M.yyyy")));

      node = node.nextSibling();
   }
}

QString ParserFQ_2::getText(const QDomNode &node)
{
   if (!node.isNull())
      if (node.nodeType() == QDomNode::TextNode)
         return node.toText().data();
   return QString();
}

QString ParserFQ_2::removeStyle(QString html)
{
   HtmlParser *htmlParser = HtmlParser::htmlParser();
   int iterator = 0;
   int tagPosition = 0;
   int tagLength = 0;

   if (htmlParser->findTag(html, 0, "body", tagPosition, tagLength))
      html = htmlParser->contentsOfTag(html, 0, "body");

   iterator = 0;
   while(htmlParser->findTag(html, iterator, "p", tagPosition, tagLength))
   {
      html.remove(tagPosition, tagLength);
      html.insert(tagPosition, "<p>");
      iterator = tagPosition + 3;
   }

   iterator = 0;
   while(htmlParser->findTag(html, iterator, "span", tagPosition, tagLength))
   {
      html.remove(tagPosition, tagLength);
      iterator = tagPosition;
   }

   iterator = 0;
   while(htmlParser->findTag(html, iterator, "/span", tagPosition, tagLength))
   {
      html.remove(tagPosition, tagLength);
      iterator = tagPosition;
   }

   return html;
}

QMap<QString, QString> ParserFQ_2::substitudeNewNamesForImages(QString &html)
{
   QMap<QString, QString> values;
   
   int iterator = 0;
   QString src;
   while(HtmlParser::htmlParser()->getTagAttribute(html, iterator, "img", "src", src))
      values[src] = Random::random() -> getARandomName();

   QMap<QString, QString>::iterator i = values.begin();
   for (;i != values.end(); i++)
   {
      QRegExp regExp(QString("\\Wsrc\\s*=\\s*\"%1\"").arg(i.key()), Qt::CaseInsensitive);
      if (regExp.indexIn(html) != -1)
      {
         html.remove(regExp.pos(), regExp.matchedLength());
         html.insert(regExp.pos(), QString(" id=\"%1\" ").arg(i.value()));
      }
   }

   return values;
}

//======================================================================

bool ParserFQ_3::read(QDomDocument *doc, CardBoxes *cardBoxes)
{
   QDomElement root = doc -> documentElement();
   CardBoxesInformation information;
   information.setAuthor(root.attribute("author"));
   information.setAuthorEmail(root.attribute("authoremail"));
   information.setLicense(root.attribute("license"));
   information.setComment(root.attribute("comment"));
   cardBoxes -> setInformation(information);
   QDomNode node = root.firstChild();
   while (!node.isNull())
   {
      if (node.toElement().tagName() == "box")
      {
         CardBox *box = new CardBox(0);
         parseBox(node.toElement(), box);
         cardBoxes->append(box);
      }
      node = node.nextSibling();
   }
   return true;
}

void ParserFQ_3::parseBox(const QDomElement &element, CardBox *box)
{
   box->setBoxName(element.attribute("name"));
   QDomNode node = element.firstChild();
   int stage = 0;
   while (!node.isNull())
   {
      BoxSection boxSection;
      if (node.toElement().tagName() == "stage")
         boxSection = BoxSection_ActiveCards;
      else if (node.toElement().tagName() == "stack")
         boxSection = BoxSection_Stack;
      else
      {
         node = node.nextSibling();
         continue;
      }

      if (boxSection == BoxSection_ActiveCards)
      {
         box->addStage();
         parseStageOrStack(node.toElement(), box, stage);
         stage++;
      }
      else
         parseStageOrStack(node.toElement(), box, -1);

      node = node.nextSibling();
   }
}

void ParserFQ_3::parseStageOrStack(const QDomElement &element, CardBox *box, int stage)
{
   QDomNode node = element.firstChild();
   while (!node.isNull())
   {
      if (node.toElement().tagName() == "card")
      {
         Card *card = 0;
         QString cardType = node.toElement().attribute("type");
         if (cardType == "regular")
            card = parseCardRegular(node.toElement());
         else if (cardType == "vocabulary")
            card = parseCardVocabulary(node.toElement());
         else if (cardType == "map")
            card = parseCardMap(node.toElement());
         
         if (card)
         {
            if (stage == -1)
               box -> addCardToStack(card, BeginningOrEnd_End);
            else
               box -> addCardToActiveCards(card, stage, BeginningOrEnd_End);
         }
      }
      node = node.nextSibling();
   }
}

Card* ParserFQ_3::parseCardVocabulary(const QDomElement &element)
{
   CardVocabulary *card = new CardVocabulary;
   QDomNode node = element.firstChild();
   while (!node.isNull())
   {
      QString tagName = node.toElement().tagName();
      if (tagName == "translationdocument")
      {
         card -> getTranslationDocuments() << new TranslationDocument(node.toElement().attribute("language"));
         parseDocument(node.toElement(), card->getTranslationDocuments().last());
      }
      else if (tagName == "statistics")
         parseCardStatistics(node.toElement(), card->getCardStatistics());
      else if (tagName == "examples")
         card -> setExamples(getText(node.firstChild()));
      else if (tagName == "wordtype")
      {
         int wordType = 0;
         QStringList list = getText(node.firstChild()).split(QRegExp("\\s*,\\s*"));
         for (int i=0; i<list.size(); i++)
         {
            if (list[i].compare("Noun", Qt::CaseInsensitive) == 0)
               wordType |= WordType_Noun;
            else if (list[i].compare("Verb", Qt::CaseInsensitive) == 0)
               wordType |= WordType_Verb;
            else if (list[i].compare("Adjective", Qt::CaseInsensitive) == 0)
               wordType |= WordType_Adjective;
            else if (list[i].compare("Adverb", Qt::CaseInsensitive) == 0)
               wordType |= WordType_Adverb;
            else if (list[i].compare("Other", Qt::CaseInsensitive) == 0)
               wordType |= WordType_Other;
         }
         card -> setWordType(wordType);
      }
      else if (tagName == "userdefinedwordtype")
         card -> setUserDefinedWordType(getText(node.firstChild()));
      else if (tagName == "comments")
         card -> setComments(getText(node.firstChild()));
      else if (tagName == "synonyms")
         card -> setSynonyms(getText(node.firstChild()).split(QRegExp("\\s*,\\s*"), QString::SkipEmptyParts));
      else if (tagName == "antonyms")
         card -> setAntonyms(getText(node.firstChild()).split(QRegExp("\\s*,\\s*"), QString::SkipEmptyParts));

      node = node.nextSibling();
   }
   return card;
}

Card* ParserFQ_3::parseCardMap(const QDomElement &element)
{
   CardMap *card = new CardMap;
   QDomNode node = element.firstChild();
   while (!node.isNull())
   {
      QString tagName = node.toElement().tagName();
      if (tagName == "map")
         card -> setMap(getText(node.firstChild()));
      else if (tagName == "region")
         card -> setRegion(getText(node.firstChild()));
      else if (tagName == "comments")
         card -> setComments(getText(node.firstChild()));
      else if (tagName == "statistics")
         parseCardStatistics(node.toElement(), card->getCardStatistics());

      node = node.nextSibling();
   }
   return card;
}

//======================================================================

bool ParserCSV::read(QTextStream *stream, CardBoxes *cardBoxes)
{
   CSVImportDialog dialog;
   if (!dialog.exec())
      return false;

   QString fieldSeparator = dialog.getFieldSeparator();
   QString textSeparator = dialog.getTextSeparator();
   CardType cardType = dialog.getCardType();
   QStringList languages = dialog.getLanguages();

   CardBox *cardBox = new CardBox(5, "CSV");

   QList<QStringList> listOfCards = splitDocument(stream->readAll(), fieldSeparator, textSeparator);
   for (int i=0; i<listOfCards.size(); i++)
   {
      QStringList list = listOfCards[i];

      Card *card = 0;
      
      if (cardType == CardType_Regular)
      {
         CardRegular *cardRegular = new CardRegular;
         cardRegular -> getFrontSideDocument() -> setText(list.value(0));
         list.removeFirst();
         cardRegular -> getBackSideDocument() -> setText(list.join(QChar('\n')));
         card = cardRegular;
      }
      else if (cardType == CardType_Vocabulary)
      {
         CardVocabulary *cardVocabulary = new CardVocabulary;
         
         cardVocabulary -> getTranslationDocuments()
            << new TranslationDocument(languages.value(0));
         cardVocabulary -> getTranslationDocuments()[0] -> setText(list.value(0));

         cardVocabulary -> getTranslationDocuments()
            << new TranslationDocument(languages.value(1));
         cardVocabulary -> getTranslationDocuments()[1] -> setText(list.value(1));

         for (int i=2; i<list.size(); i++)
         {
            cardVocabulary -> getTranslationDocuments()
               << new TranslationDocument(languages.value(i));
            cardVocabulary -> getTranslationDocuments()[i] -> setText(list.value(i));
         }

         card = cardVocabulary;
      }
      if (card)
         cardBox -> addCardToActiveCards(card, 0, BeginningOrEnd_End);
   }
   cardBoxes -> append(cardBox);
   return true;
}

QList<QStringList> ParserCSV::splitDocument(const QString &document, const QString &fieldSeparator, const QString &textSeparator)
{
   QList<QStringList> splittedDocument;
   QStringList valueList;
   int iterator = 0;
   while(iterator<document.size())
   {
      bool listFinished = false;
      valueList << extractValue(document, fieldSeparator, textSeparator, &iterator, &listFinished);

      if (listFinished)
      {
         splittedDocument << valueList;
         valueList.clear();
      }
   }
   return splittedDocument;
}

QString ParserCSV::extractValue(const QString &document, const QString &fieldSeparator,
                                QString textSeparator, int *iteratorPtr, bool *listFinished)
{
   int &iterator = *iteratorPtr;
   QString value;

   int i;
   for(i=iterator; document[i] == ' '; i++);
   if (QRegExp(QRegExp::escape(textSeparator) + ".*").indexIn(document, i) == i)
//   Due to a different behaviour of qt 4.5 from previous versions the line below should not be used
//   (replaced by the line above)
//   if (document.indexOf(QRegExp(QRegExp::escape(textSeparator) + ".*"), i) == i)
      iterator = i+textSeparator.size();
   else
      textSeparator = "";

   QRegExp regExp(QRegExp::escape(textSeparator) + "\\s*"
                  + "(" + QRegExp::escape(fieldSeparator) + "|\n" + "|$" + ")");
   int index = regExp.indexIn(document, iterator);
   if (index == -1) //should not happen
      index = document.size();
   value = document.mid(iterator, index-iterator);
   iterator = index + regExp.matchedLength();
   if (regExp.cap(1) != fieldSeparator)
      *listFinished = true;
   else
      *listFinished = false;
   
   return value;
}


//======================================================================

bool ParserKVTML_2_0::read(QDomDocument *doc, CardBoxes *cardBoxes)
{
   QDomElement root = doc -> documentElement();
   QDomNodeList nodeList = root.elementsByTagName("identifiers");
   if (nodeList.isEmpty())
   {
      somethingIsWrong();
      return false;
   }

   getLanguages(nodeList.at(0).toElement());

   //----------
   
   nodeList = root.elementsByTagName("information");
   if (nodeList.isEmpty())
   {
      somethingIsWrong();
      return false;
   }

   QDomNode node = nodeList.at(0).toElement().firstChild();
   while(!node.isNull())
   {
      QString tagName = node.toElement().tagName();
      CardBoxesInformation information;
      if (tagName == "author")
         information.setAuthor(getText(node.firstChild()));
      else if (tagName == "contact")
         information.setAuthorEmail(getText(node.firstChild()));
      else if (tagName == "license")
         information.setLicense(getText(node.firstChild()));
      else if (tagName == "comment")
         information.setComment(getText(node.firstChild()));
      cardBoxes -> setInformation(information);
      node = node.nextSibling();
   }

   //----------

   nodeList = root.elementsByTagName("entries");
   if (nodeList.isEmpty())
   {
      somethingIsWrong();
      return false;
   }

   node = nodeList.at(0).toElement().firstChild();
   while(!node.isNull())
   {
      if (node.toElement().tagName() == "entry")
      {
         Card *card = 0;
         card = parseCardVocabulary(node.toElement());
         cards << card;
      }
      node = node.nextSibling();
   }

   //----------

   nodeList = root.elementsByTagName("lessons");
   if (nodeList.isEmpty())
   {
      somethingIsWrong();
      return false;
   }

   node = nodeList.at(0).toElement().firstChild();
   while(!node.isNull())
   {
      if (node.toElement().tagName() == "container")
      {
         CardBox *box = 0;
         box = parseCardBox(node.toElement());
         cardBoxes -> append(box);
      }
      node = node.nextSibling();
   }

   return true;
}

void ParserKVTML_2_0::getLanguages(const QDomElement &element)
{
   QDomNode node = element.firstChild();
   while(!node.isNull())
   {
      QString tagName = node.toElement().tagName();
      if (tagName == "identifier")
      {
         QDomNode innerNode = node.toElement().firstChild();
         while(!innerNode.isNull())
         {
            QString innerTagName = innerNode.toElement().tagName();
            if (innerTagName == "name")
               languages << getText(innerNode.firstChild());
            innerNode = innerNode.nextSibling();
         }
      }
      node = node.nextSibling();
   }
}

Card* ParserKVTML_2_0::parseCardVocabulary(const QDomElement &element)
{
   CardVocabulary *card = new CardVocabulary;
   int translationId = 0;
   QDomNode node = element.firstChild();
   while(!node.isNull())
   {
      QString tagName = node.toElement().tagName();
      if (tagName == "translation")
      {
         card -> getTranslationDocuments() << new TranslationDocument();
         parseTranslation(node.toElement(), card, card->getTranslationDocuments().last());
         translationId++;
      }

      node = node.nextSibling();
   }
   return card;
}

QString ParserKVTML_2_0::getText(const QDomNode &node)
{
   if (!node.isNull())
      if (node.nodeType() == QDomNode::TextNode)
         return node.toText().data();
   return QString();
}

void ParserKVTML_2_0::somethingIsWrong()
{
      QMessageBox::warning(0, QObject::tr("Parser"),
                           QObject::tr("Program encountered an error while reading the file."));
}

void ParserKVTML_2_0::parseTranslation(const QDomElement &element, CardVocabulary *card, TranslationDocument *translationDocument)
{
   translationDocument -> setLanguageName(languages[element.attribute("id", "0").toInt()]);
   QString text;
   QString examples;
   QString comments;
   QString imagePath;
   QDomNode node = element.firstChild();
   while (!node.isNull())
   {
      QString tagName = node.toElement().tagName();
      if (tagName == "text")
         text = getText(node.firstChild());
      else if (tagName == "example")
         examples += getText(node.firstChild()) + "\n";
      else if (tagName == "comment")
         comments += getText(node.firstChild()) + "\n";
      else if (tagName == "image")
         imagePath = QUrl::fromEncoded(QByteArray(getText(node.firstChild()).toAscii().constData())).path();

      node = node.nextSibling();
   }

   if (!imagePath.isEmpty())
   {
      QFile imageFile(imagePath);
      if (imageFile.open(QIODevice::ReadOnly))
      {
         QString imageName = Random::random() -> getARandomName();
         translationDocument -> addAttachment(imageFile.readAll(), imageName, ATTACHMENT_IMAGE);
         
         text += QString("<br><img id=\"%1\" />").arg(imageName);
      }
   }

   translationDocument -> setText(text);
   card -> setExamples(card->getExamples() + examples);
   card -> setComments(card->getComments() + comments);
}

CardBox* ParserKVTML_2_0::parseCardBox(const QDomElement &element)
{
   CardBox *box = new CardBox(5);
   QDomNode node = element.firstChild();
   while(!node.isNull())
   {
      if (node.toElement().tagName() == "name")
         box -> setBoxName(getText(node.firstChild()));
      else if ( node.toElement().tagName() == "entry")
      {
         int cardNumber = node.toElement().attribute("id").toInt();
         box -> addCardToActiveCards(cards[cardNumber], 0, BeginningOrEnd_End);
      }

      node = node.nextSibling();
   }
   return box;
}

