// +-------------------------------------------------------------------------+
// | FlashQard - An efficient way to memorize.                               |
// | Copyright (C) 2008-2009 Shahab <shahab@flashqard-project.org>           |
// +-------------------------------------------------------------------------+
// | License:                                                                |
// |                                                                         |
// | FlashQard is free software; you can redistribute it and/or modify       |
// | it under the terms of the GNU General Public License as published by    |
// | the Free Software Foundation; either version 2 of the License, or       |
// | (at your option) any later version.                                     |
// |                                                                         |
// | FlashQard is distributed in the hope that it will be useful,            |
// | but WITHOUT ANY WARRANTY; without even the implied warranty of          |
// | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the            |
// | GNU General Public License for more details.                            |
// |                                                                         |
// | You should have received a copy of the GNU General Public License       |
// | along with FlashQard; if not, write to the Free Software                |
// | Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA|
// |                                                                         |
// +-------------------------------------------------------------------------+
// | Author:  Shahab Shirazi <shahab@flashqard-project.org>                  |
// +-------------------------------------------------------------------------+

#include "FileWriter.h"
#include "FlashQard.h"
#include "CardRegular.h"
#include "CardVocabulary.h"
#include "CardMap.h"
#include "Utilities.h"
#include "CSVExportDialog.h"
#include <QString>
#include <QFile>
#include <QMessageBox>
#include <QtXml>
#include <QApplication>
#include <QTime>
#include <stdexcept>
using std::runtime_error;

FileWriterSettings::FileWriterSettings()
{
   stageStackMode = StageStack_DoNotMove;
   stageNumber = 0;
   exportStatistics = true;
   sortCards = false;
   sortBy = SortBy_CreationDate;
   sortOrder = SortOrder_Ascending;
   fileFormat = FileFormat_FQ;
}

//======================================================================

bool FileWriter::writeToFile(const QString &fileName, const CardBoxes *cardBoxes, const FileWriterSettings &config)
{
   //We make a temporary file, then if writing was successful
   //we rename the temporary file otherwise we delete it
   QFileInfo fileInfo(fileName);
   QString tempFileName;
#ifdef Q_WS_WIN
   tempFileName = "Temporary_File_" + QTime::currentTime().toString("mm_ss_zzz");
#else
   tempFileName = ".Temporary_File_" + QTime::currentTime().toString("mm_ss_zzz");
#endif
   QFile tempFile (fileInfo.dir().path() + QDir::separator() + tempFileName);
   if (!tempFile.open(QIODevice::WriteOnly))
   {
      QMessageBox::warning(0, QObject::tr("Save File"),
                           QObject::tr("Unable to create a temporary file: %1").arg(tempFileName));
      return false;
   }

   CardBoxes newCardBoxes;
   addCardsToTheNewCardBoxes(cardBoxes, &newCardBoxes, config);

   bool successfullySaved = true;
   try
   {
      if (fileInfo.suffix().compare("csv", Qt::CaseInsensitive) == 0)
      {
         FileWriterCSV csvWriter;
         csvWriter.write(&tempFile, &newCardBoxes);
      }
      else
      {
         FileWriterFQ fqWriter;
         fqWriter.write(&tempFile, &newCardBoxes);
      }
   }
   catch (runtime_error &error)
   {
      QString errorMessage = error.what();
      if (!errorMessage.isEmpty())
      {
         QMessageBox::warning(0, QObject::tr("Save File"),
                 QObject::tr("Saving failed:\n%1").arg(errorMessage));
      }
      successfullySaved = false;
   }

   bool successfullyRenamed = false;
   if (successfullySaved)
   {
      if (fileInfo.exists())
         QFile::remove(fileInfo.filePath());
      if (!(successfullyRenamed = tempFile.rename(fileInfo.filePath())))
      {
         QMessageBox::warning(0, QObject::tr("Save File"),
                              QObject::tr("Unable to rename temporary file."));
      }
   }

   if (!successfullyRenamed || !successfullySaved)
   {
      tempFile.remove();
      return false;
   }
   else
      return true;

}

void FileWriter::addCardsToTheNewCardBoxes(const CardBoxes *oldCardBoxes,
                                           CardBoxes *newCardBoxes,
                                           const FileWriterSettings &config)
{
   if (config.fileFormat == FileWriterSettings::FileFormat_FQ)
   {
      for (int newCategory=0, oldCategory=0; oldCategory<oldCardBoxes -> size(); oldCategory++)
      {
         //config.category.size()==0 means all categories
         if (!config.categories.contains(oldCategory) && config.categories.size() > 0)
            continue;

         newCardBoxes -> append(new CardBox(oldCardBoxes->at(oldCategory)->countStages(), oldCardBoxes->at(oldCategory)->getBoxName()));
      
         CardPosition pos = oldCardBoxes -> at(oldCategory) -> firstPosition();
         while(oldCardBoxes->at(oldCategory)->isPositionValid(pos))
         {
            Card *card = cloneCard(oldCardBoxes->at(oldCategory)->at(pos), config.exportStatistics);

            if (config.stageStackMode == FileWriterSettings::StageStack_MoveToStack)
               newCardBoxes -> at(newCategory) -> addCardToStack(card, BeginningOrEnd_End);
            else if (config.stageStackMode == FileWriterSettings::StageStack_MoveToStage)
            {
               while(config.stageNumber >= newCardBoxes -> at(newCategory) -> countStages())
                  newCardBoxes -> at(newCategory) -> addStage();
               newCardBoxes -> at(newCategory) -> addCardToActiveCards(card, config.stageNumber, BeginningOrEnd_End);
            }
            else //StageStack_DoNotMove
            {
               //enough stages have been added in advance
//            while(oldCardBoxes->at(oldCategory)->countStages() > newCardBoxes -> at(newCategory) -> countStages())
//               newCardBoxes -> at(newCategory) -> addStage();
               if (pos.boxSection == BoxSection_Stack)
                  newCardBoxes -> at(newCategory) -> addCardToStack(card, BeginningOrEnd_End);
               else
                  newCardBoxes -> at(newCategory) -> addCardToActiveCards(card, pos.stage, BeginningOrEnd_End);
            }
               
            pos = oldCardBoxes -> at(oldCategory) -> nextPosition(pos, MovePolicy_ThroughStackAndActiveCards);
         }
      
         if(config.sortCards)
            newCardBoxes -> at(newCategory) -> sortAllCards(config.sortBy, config.sortOrder);
      
         newCategory++;
      }
   }
   else // config.fileFormat == FileWriterSettings::FileFormat_CSV
   {
      // put all cards into stack then sort them
      // (when saving csv files, information about stack, stages and cardBoxes will not be saved)
      
      newCardBoxes -> append(new CardBox(0, "CSV"));
      for (int i=0; i<oldCardBoxes->size(); i++)
      {
         CardPosition pos = oldCardBoxes -> at(i) -> firstPosition();
         while(oldCardBoxes->at(i)->isPositionValid(pos))
         {
            Card *card = cloneCard(oldCardBoxes->at(i)->at(pos), false);
            newCardBoxes -> at(0) -> addCardToStack(card, BeginningOrEnd_End);
            pos = oldCardBoxes -> at(i) -> nextPosition(pos, MovePolicy_ThroughStackAndActiveCards);
         }
      }
      newCardBoxes -> at(0) -> sortStack(config.sortBy, config.sortOrder);
   }

   newCardBoxes -> setInformation(oldCardBoxes -> getInformation());
}

Card* FileWriter::cloneCard(const Card *card, bool importStatistics) const
{
   Card *newCard = card -> clone();
   if (!importStatistics)
      newCard -> clearAnswers();
   return newCard;
}

//======================================================================

void FileWriterFQ::write(QIODevice *device, const CardBoxes *cardBoxes)
{
   const int Indent = 1;
   QDomDocument doc;
   QDomElement root = doc.createElement("flashcards");
   root.setAttribute("writerversion", 3);
   root.setAttribute("flashqardversion", VERSION);
   root.setAttribute("author", cardBoxes->getInformation().getAuthor());
   root.setAttribute("authoremail", cardBoxes->getInformation().getAuthorEmail());
   root.setAttribute("license", cardBoxes->getInformation().getLicense());
   root.setAttribute("comment", cardBoxes->getInformation().getComment());
   for (int i=0; i<cardBoxes->size(); i++)
   {
      QDomElement box = doc.createElement("box");
      box.setAttribute("name", cardBoxes->at(i)->getBoxName());

      QDomElement stack = doc.createElement("stack");
      for (int s=0; s<cardBoxes->at(i)->countStack(); s++)
      {
         const Card *cardPointer = cardBoxes->at(i)->at(CardPosition(s));
         QDomElement cardElement = doc.createElement("card");
         writeCard(doc, cardElement, cardPointer);
         stack.appendChild(cardElement);
      }
      box.appendChild(stack);

      for (int j=0; j<cardBoxes->at(i)->countStages(); j++)
      {
         QDomElement stage = doc.createElement("stage");
         for (int k=0; k<cardBoxes->at(i)->countActiveCards(j); k++)
         {
            const Card *cardPointer = cardBoxes->at(i)->at(CardPosition(j, k));
            QDomElement cardElement = doc.createElement("card");
            writeCard(doc, cardElement, cardPointer);

            stage.appendChild(cardElement);
         }
         box.appendChild(stage);
      }
      root.appendChild(box);
   }
   doc.appendChild(root);

   QTextStream out(device);
   doc.save(out, Indent);

}

void FileWriterFQ::writeCard(QDomDocument &doc, QDomElement &cardElement, const Card *cardPointer)
{
   switch(cardPointer->getType())
   {
      case CardType_Regular:
         writeCardRegular(doc, cardElement, static_cast<const CardRegular*>(cardPointer));
         break;

      case CardType_Vocabulary:
         writeCardVocabulary(doc, cardElement, static_cast<const CardVocabulary*>(cardPointer));
         break;

      case CardType_Map:
         writeCardMap(doc, cardElement, static_cast<const CardMap*>(cardPointer));
         break;
   }
}

void FileWriterFQ::writeCardRegular(QDomDocument &doc, QDomElement &cardElement, const CardRegular *cardPointer)
{
   cardElement.setAttribute("type", "regular");

   QDomElement frontSideElement = doc.createElement("frontsidedocument");
   writeDocument(doc, frontSideElement, cardPointer->getFrontSideDocument());
   cardElement.appendChild(frontSideElement);

   QDomElement backSideElement = doc.createElement("backsidedocument");
   writeDocument(doc, backSideElement, cardPointer->getBackSideDocument());
   cardElement.appendChild(backSideElement);

   QDomElement statisticsElement = doc.createElement("statistics");
   writeCardStatistics(doc, statisticsElement, cardPointer->getCardStatistics());
   cardElement.appendChild(statisticsElement);

   QDomElement commentsElement = doc.createElement("comments");
   QDomText commentsText = doc.createTextNode(cardPointer->getComments());
   commentsElement.appendChild(commentsText);
   cardElement.appendChild(commentsElement);

}

void FileWriterFQ::writeCardVocabulary(QDomDocument &doc, QDomElement &cardElement, const CardVocabulary *cardPointer)
{
   cardElement.setAttribute("type", "vocabulary");
   for (int i=0; i<cardPointer->getTranslationDocuments().size(); i++)
   {

      QDomElement translationElement = doc.createElement("translationdocument");
      translationElement.setAttribute("language", cardPointer->getTranslationDocuments()[i]->getLanguageName());
      writeDocument(doc, translationElement, cardPointer->getTranslationDocuments()[i]);
      cardElement.appendChild(translationElement);
   }

   QDomElement examplesElement = doc.createElement("examples");
   QDomText examplesText = doc.createTextNode(cardPointer->getExamples());
   examplesElement.appendChild(examplesText);
   cardElement.appendChild(examplesElement);

   QDomElement wordTypeElement = doc.createElement("wordtype");
   int wordType = cardPointer->getWordType();
   QStringList wordTypes;
   if (wordType & WordType_Noun)
      wordTypes << "Noun";
   if (wordType & WordType_Verb)
      wordTypes << "Verb";
   if (wordType & WordType_Adjective)
      wordTypes << "Adjective";
   if (wordType & WordType_Adverb)
      wordTypes << "Adverb";
   if (wordType & WordType_Other)
      wordTypes << "Other";
   QDomText wordTypeText = doc.createTextNode(wordTypes.join(", "));
   wordTypeElement.appendChild(wordTypeText);
   cardElement.appendChild(wordTypeElement);

   if (wordType & WordType_Other)
   {
      QDomElement userDefinedWordTypeElement = doc.createElement("userdefinedwordtype");
      QDomText userDefinedWordTypeText = doc.createTextNode(cardPointer->getUserDefinedWordType());
      userDefinedWordTypeElement.appendChild(userDefinedWordTypeText);
      cardElement.appendChild(userDefinedWordTypeElement);
   }

   QDomElement commentsElement = doc.createElement("comments");
   QDomText commentsText = doc.createTextNode(cardPointer->getComments());
   commentsElement.appendChild(commentsText);
   cardElement.appendChild(commentsElement);

   QDomElement synonymsElement = doc.createElement("synonyms");
   QDomText synonymsText = doc.createTextNode(cardPointer->getSynonyms().join(", "));
   synonymsElement.appendChild(synonymsText);
   cardElement.appendChild(synonymsElement);

   QDomElement antonymsElement = doc.createElement("antonyms");
   QDomText antonymsText = doc.createTextNode(cardPointer->getAntonyms().join(", "));
   antonymsElement.appendChild(antonymsText);
   cardElement.appendChild(antonymsElement);

   QDomElement statisticsElement = doc.createElement("statistics");
   writeCardStatistics(doc, statisticsElement, cardPointer->getCardStatistics());
   cardElement.appendChild(statisticsElement);
}

void FileWriterFQ::writeCardMap(QDomDocument &doc, QDomElement &cardElement, const CardMap *cardPointer)
{
   cardElement.setAttribute("type", "map");

   QDomElement mapElement = doc.createElement("map");
   QDomText mapText = doc.createTextNode(cardPointer->getMap());
   mapElement.appendChild(mapText);
   cardElement.appendChild(mapElement);

   QDomElement regionElement = doc.createElement("region");
   QDomText regionText = doc.createTextNode(cardPointer->getRegion());
   regionElement.appendChild(regionText);
   cardElement.appendChild(regionElement);

   QDomElement commentsElement = doc.createElement("comments");
   QDomText commentsText = doc.createTextNode(cardPointer->getComments());
   commentsElement.appendChild(commentsText);
   cardElement.appendChild(commentsElement);

   QDomElement statisticsElement = doc.createElement("statistics");
   writeCardStatistics(doc, statisticsElement, cardPointer->getCardStatistics());
   cardElement.appendChild(statisticsElement);
}

void FileWriterFQ::writeDocument(QDomDocument &doc, QDomElement &documentElement, const Document *document)
{
   QDomElement html = doc.createElement("html");
   QDomText htmlText = doc.createTextNode(document->getText());
   html.appendChild(htmlText);
   documentElement.appendChild(html);

   for (int i=0; i<document->attachmentCount(); i++)
   {
      QDomElement attachment = doc.createElement("attachment");
      attachment.setAttribute("name", document->getAttachmentName(i));
      attachment.setAttribute("type", document->getAttachmentType(i));
      QDomText attachmentText = doc.createTextNode(document->attachmentToBase64(i));
      attachment.appendChild(attachmentText);
      documentElement.appendChild(attachment);
   }
}

void FileWriterFQ::writeCardStatistics(QDomDocument &doc, QDomElement &statisticsElement, const CardStatistics &statistics)
{
   QDomElement dateCreated = doc.createElement("dateCreated");
   QDomText dateCreatedText = doc.createTextNode(statistics.getDateCreated().toString("d.M.yyyy"));
   dateCreated.appendChild(dateCreatedText);
   statisticsElement.appendChild(dateCreated);

   for (int i=0; i<statistics.getAnswers().size(); i++)
   {
      QDomElement answer = doc.createElement("answer");
      answer.setAttribute("date", statistics.getAnswers()[i].getDate().toString("d.M.yyyy"));
      QDomText answerText = doc.createTextNode(statistics.getAnswers()[i].isCorrect() ? "true" : "false");
      answer.appendChild(answerText);
      statisticsElement.appendChild(answer);
   }

}


//======================================================================


void FileWriterCSV::write(QIODevice *device, const CardBoxes *cardBoxes)
{
   CSVExportDialog dialog;
   if (!dialog.exec())
      throw runtime_error("");

   QString fieldSeparator = dialog.getFieldSeparator();
   QString textSeparator = dialog.getTextSeparator();
   QTextStream out(device);

   for (int box=0; box<cardBoxes->size(); box++)
   {
      CardPosition position = cardBoxes->at(box)->firstPosition();
      while(cardBoxes->at(box)->isPositionValid(position))
      {
         const Card *cardPointer = cardBoxes->at(box)->at(position);
         QString line;
         switch(cardPointer->getType())
         {
            case CardType_Regular:
               line = cardRegularToString(static_cast<const CardRegular*>(cardPointer), fieldSeparator, textSeparator);
               break;

            case CardType_Vocabulary:
               line = cardVocabularyToString(static_cast<const CardVocabulary*>(cardPointer), fieldSeparator, textSeparator);
               break;
         }
         out << line << "\n";

         position = cardBoxes -> at(box) -> nextPosition(position, MovePolicy_ThroughStackAndActiveCards);
      }
   }
}

QString FileWriterCSV::cardRegularToString(const CardRegular *card,
                                           const QString &fieldSeparator,
                                           const QString &textSeparator) const
{
   HtmlParser *htmlParser = HtmlParser::htmlParser();
   QString line = textSeparator
                + htmlParser->htmlToPlainText(card->getFrontSideDocument()->getText())
                + textSeparator
                + fieldSeparator
                + textSeparator
                + htmlParser->htmlToPlainText(card->getBackSideDocument()->getText())
                + textSeparator;
   return line;
}

QString FileWriterCSV::cardVocabularyToString(const CardVocabulary *card,
                                              const QString &fieldSeparator,
                                              const QString &textSeparator) const
{
   QString line;
   HtmlParser *htmlParser = HtmlParser::htmlParser();

   for (int i=0; i<card->getTranslationDocuments().size(); i++)
   {
      if (i!=0)
         line += fieldSeparator;
      line += textSeparator
            + htmlParser->htmlToPlainText(card->getTranslationDocuments()[i]->getText())
            + textSeparator;
   }
   return line;
}


