// +-------------------------------------------------------------------------+
// | FlashQard - An efficient way to memorize.                               |
// | Copyright (C) 2008-2009 Shahab <shahab@flashqard-project.org>           |
// +-------------------------------------------------------------------------+
// | License:                                                                |
// |                                                                         |
// | FlashQard is free software; you can redistribute it and/or modify       |
// | it under the terms of the GNU General Public License as published by    |
// | the Free Software Foundation; either version 2 of the License, or       |
// | (at your option) any later version.                                     |
// |                                                                         |
// | FlashQard is distributed in the hope that it will be useful,            |
// | but WITHOUT ANY WARRANTY; without even the implied warranty of          |
// | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the            |
// | GNU General Public License for more details.                            |
// |                                                                         |
// | You should have received a copy of the GNU General Public License       |
// | along with FlashQard; if not, write to the Free Software                |
// | Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA|
// |                                                                         |
// +-------------------------------------------------------------------------+
// | Author:  Shahab Shirazi <shahab@flashqard-project.org>                  |
// +-------------------------------------------------------------------------+

#include "PrettyButton.h"
#include "TextEditor.h"
#include "MimeData.h"
#include <QPainter>
#include <QTimerEvent>
#include <QDropEvent>

#define INTERVALTIME 50

PrettyButton::PrettyButton(Qt::Orientation o, QWidget *parent)
   : QAbstractButton(parent)
{
   setOrientation(o);
   setMouseTracking(true);
   setAcceptDrops(true);
   buttonState = ButtonState_Normal;
   timerId = 0;
   normalColor = QColor(palette().color(QPalette::Active, QPalette::Button));
   hoveredColor = QColor(57, 102, 178);
   hoveredPressedColor = QColor(80, 140, 250);
   pressedColor = QColor(12, 54, 177);
   animationPercentage = 0;
   animationTime = 500;
   connect(this, SIGNAL(toggled(bool)), this, SLOT(toggledSlot()));
}

void PrettyButton::setOrientation(Qt::Orientation o)
{
   orientation = o;
   if (orientation == Qt::Horizontal)
      setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Preferred);
   else
      setSizePolicy(QSizePolicy::Preferred, QSizePolicy::Expanding);
}

Qt::Orientation PrettyButton::getOrientation() const
{
   return orientation;
}

void PrettyButton::setAnimationTime(int at)
{
   animationTime = at;
}

int PrettyButton::getAnimationTime() const
{
   return animationTime;
}

QSize PrettyButton::sizeHint() const
{
   QSize size;
   size = QFontMetrics(font()).boundingRect(text()).size();
   size.setWidth(size.width()+10 + iconSize().width());
   size.setHeight(size.height()+5);
   if (orientation == Qt::Horizontal)
      return size;
   else
      return QSize(size.height(), size.width());
}

void PrettyButton::paintEvent(QPaintEvent*)
{
   QPixmap pixmap(orientation==Qt::Horizontal ? width() : height(),
                  orientation==Qt::Horizontal ? height() : width());
   pixmap.fill(Qt::transparent);
   QPainter pixmapPainter(&pixmap);
   pixmapPainter.setRenderHint(QPainter::Antialiasing);

   QPainterPath path;
   path.addRoundedRect(0, 0, pixmap.width(), pixmap.height(), 7, 7);
   QColor color;
   if (isEnabled())
   {
      QColor finalColor = isDown() || isChecked() ? pressedColor : hoveredColor;
      QColor initialColor = isDown() || isChecked() ? hoveredPressedColor : normalColor;
      color.setRed( (finalColor.red()-initialColor.red())*animationPercentage/100.0 + initialColor.red());
      color.setGreen( (finalColor.green()-initialColor.green())*animationPercentage/100.0 + initialColor.green());
      color.setBlue( (finalColor.blue()-initialColor.blue())*animationPercentage/100.0 + initialColor.blue());
      pixmapPainter.setPen(QColor(255*animationPercentage/100, 255*animationPercentage/100, 255*animationPercentage/100));
   }
   else
   {
      color = normalColor;
      pixmapPainter.setPen(palette().color(QPalette::Disabled, QPalette::ButtonText));
   }
   
   QLinearGradient gradient(0, 0, 0, pixmap.height());
   QColor lighterColor = color;
   lighterColor.setRgb(color.red()<=225 ? color.red()+30 : 255,
                       color.green()<=225 ? color.green()+30 : 255,
                       color.blue()<=225 ? color.blue()+30 : 255);

   QColor darkerColor = color;
   darkerColor.setRgb(color.red()>=30 ? color.red()-30 : 0,
                       color.green()>=30 ? color.green()-30 : 0,
                       color.blue()>=30 ? color.blue()-30 : 0);

   gradient.setColorAt(0, lighterColor);
   gradient.setColorAt(0.2, color);
   gradient.setColorAt(0.8, color);
   gradient.setColorAt(1, lighterColor);
   QBrush brush(gradient);
   pixmapPainter.fillPath(path, brush);

   QPen pen(QBrush(darkerColor), 1);
   pixmapPainter.strokePath(path, pen);
   
   QSize textSize = pixmapPainter.fontMetrics().boundingRect(text()).size();
   QPixmap iconPixmap = icon().pixmap(iconSize(), isEnabled() ? QIcon::Normal : QIcon::Disabled);
   int startPosition = (pixmap.width()-iconPixmap.size().width() -textSize.width())/2;
   pixmapPainter.drawText(startPosition+iconPixmap.size().width()+5, (pixmap.height()-textSize.height())/2,
                    textSize.width(), textSize.height(),
                    Qt::AlignVCenter | Qt::TextShowMnemonic, text());

   if (orientation == Qt::Horizontal)
      pixmapPainter.drawPixmap(startPosition, (pixmap.height()-iconPixmap.size().height())/2, iconPixmap);
   else
   {
      pixmapPainter.end();
      QMatrix matrix;
      matrix.rotate(-90);
      pixmap = pixmap.transformed(matrix);

      pixmapPainter.begin(&pixmap);
      startPosition = pixmap.height()-startPosition-iconPixmap.size().height();
      pixmapPainter.drawPixmap((pixmap.width()-iconPixmap.size().width())/2, startPosition, iconPixmap);
   }

   QPainter widgetPainter(this);
   widgetPainter.drawPixmap(0,0, pixmap);
}

void PrettyButton::enterEvent(QEvent *event)
{
   QAbstractButton::enterEvent(event);
   mouseEntered();
   if (isEnabled())
   {
      emit hovered_notDragging();
      emit hovered();
   }
}

void PrettyButton::leaveEvent(QEvent *event)
{
   QAbstractButton::leaveEvent(event);
   mouseLeft();
}

void PrettyButton::mouseEntered()
{
   killTimer(timerId);
   buttonState = ButtonState_Hovered;
   timerId = startTimer(INTERVALTIME);
}

void PrettyButton::mouseLeft()
{
   killTimer(timerId);
   buttonState = ButtonState_Normal;
   timerId = startTimer(INTERVALTIME);
}

void PrettyButton::dragEnterEvent(QDragEnterEvent *event)
{
   if (event->mimeData()->hasFormat(CARD_POSITION))
      event->acceptProposedAction();   

   mouseEntered();
   if (isEnabled())
   {
      emit hovered_dragging();
      emit hovered();
   }
}

void PrettyButton::dragLeaveEvent(QDragLeaveEvent *event)
{
   QAbstractButton::dragLeaveEvent(event);
   mouseLeft();
}

void PrettyButton::mousePressEvent(QMouseEvent *event)
{
   QAbstractButton::mousePressEvent(event);

   killTimer(timerId);
   animationPercentage = 100;
   update();
}
   
void PrettyButton::mouseReleaseEvent(QMouseEvent *event)
{
   QAbstractButton::mouseReleaseEvent(event);

   killTimer(timerId);
   buttonState = ButtonState_Normal;
   timerId = startTimer(INTERVALTIME);
}

void PrettyButton::toggledSlot()
{
   killTimer(timerId);
   buttonState = ButtonState_Normal;
   if (isChecked())
      animationPercentage = 100;
   timerId = startTimer(INTERVALTIME);
}

void PrettyButton::timerEvent(QTimerEvent *event)
{
   if (event->timerId() != timerId)
      return;

   if (buttonState == ButtonState_Hovered)
   {
      if (isChecked())
         animationPercentage -= INTERVALTIME *100.0 / animationTime;
      else
         animationPercentage += INTERVALTIME *100.0 / animationTime;
   }
   else if (buttonState == ButtonState_Normal)
   {
      if (isChecked())
         animationPercentage += INTERVALTIME *100.0 / animationTime;
      else
         animationPercentage -= INTERVALTIME *100.0 / animationTime;
   }

   if (animationPercentage > 100)
   {
      animationPercentage = 100;
      killTimer(timerId);
   }
   else if (animationPercentage < 0)
   {
      animationPercentage = 0;
      killTimer(timerId);
   }
   
   update();
}

void PrettyButton::dropEvent(QDropEvent *event)
{
   const MimeData *mime = dynamic_cast<const MimeData*>(event->mimeData());
   if (mime)
      emit dropMimeDataSignal(mime, event->dropAction());
}

bool PrettyButton::isAnimating() const
{
   return (animationPercentage > 0 && animationPercentage < 100);
}
