// +-------------------------------------------------------------------------+
// | FlashQard - An efficient way to memorize.                               |
// | Copyright (C) 2008-2009 Shahab <shahab@flashqard-project.org>           |
// +-------------------------------------------------------------------------+
// | License:                                                                |
// |                                                                         |
// | FlashQard is free software; you can redistribute it and/or modify       |
// | it under the terms of the GNU General Public License as published by    |
// | the Free Software Foundation; either version 2 of the License, or       |
// | (at your option) any later version.                                     |
// |                                                                         |
// | FlashQard is distributed in the hope that it will be useful,            |
// | but WITHOUT ANY WARRANTY; without even the implied warranty of          |
// | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the            |
// | GNU General Public License for more details.                            |
// |                                                                         |
// | You should have received a copy of the GNU General Public License       |
// | along with FlashQard; if not, write to the Free Software                |
// | Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA|
// |                                                                         |
// +-------------------------------------------------------------------------+
// | Author:  Shahab Shirazi <shahab@flashqard-project.org>                  |
// +-------------------------------------------------------------------------+

#include <QtGui>

#include <QHttp>
#include "TextEditor.h"
#include "Card.h"
#include "Settings.h"
#include "Utilities.h"
#include "Downloader.h"

ObjectsForJavaScript::ObjectsForJavaScript(TextEditor *_textEditor)
   : QObject(_textEditor), textEditor(_textEditor)
{
}

QString ObjectsForJavaScript::getCss()
{
   return Settings::settings() -> makeCss();
}

QString ObjectsForJavaScript::getARandomName()
{
   QString name;
   return Random::random()->getARandomName();
}

QString ObjectsForJavaScript::getObjectPath(const QString &id)
{
   TextEditorObjectImage *imageObject = dynamic_cast<TextEditorObjectImage*>(textEditor -> findObject(id));
   if (imageObject)
      return imageObject -> getFilePath();
   else
      return QString();
}

void ObjectsForJavaScript::clickedOnObject(const QString &id)
{
   emit clickedOnObjectSignal(id);
}

void ObjectsForJavaScript::clickedOnNonObject()
{
   emit clickedOnNonObjectSignal();
}

//----------------------------------------------------------------------

WebView::WebView(QWidget *parent)
   : QWebView(parent)
{
   settings() -> setFontSize(QWebSettings::DefaultFontSize, 14);

   setAcceptDrops(true);
}

void WebView::startEditing()
{
   //NOTE: it needs a mouse click!
   //any better way to make the cursor visible?
   this -> setFocus();
   QMouseEvent mouseEventPress(QEvent::MouseButtonPress, QPoint(10,10), Qt::LeftButton, Qt::NoButton, Qt::NoModifier);
   QApplication::sendEvent(this, &mouseEventPress);
   QTimer::singleShot(50, this, SLOT(sendMouseReleaseEvent()));
}

void WebView::sendMouseReleaseEvent()
{
   QMouseEvent mouseEventRelease(QEvent::MouseButtonRelease, QPoint(10,10), Qt::LeftButton, Qt::NoButton, Qt::NoModifier);
   QApplication::sendEvent(this, &mouseEventRelease);
   pageAction(QWebPage::MoveToEndOfDocument) -> trigger();
}

void WebView::focusOutEvent(QFocusEvent *event)
{
   QWebView::focusOutEvent(event);
   emit focusOutSignal();
}

void WebView::focusInEvent(QFocusEvent *event)
{
   QWebView::focusInEvent(event);
   //NOTE: enabling the codes below will correct the focus behaviours
   //But will make some problems with font size...
//    QMouseEvent mouseEventPress(QEvent::MouseButtonPress, QPoint(1,1), Qt::LeftButton, Qt::NoButton, Qt::NoModifier);
//    QMouseEvent mouseEventRelease(QEvent::MouseButtonRelease, QPoint(1,1), Qt::LeftButton, Qt::NoButton, Qt::NoModifier);
//    QApplication::sendEvent(this, &mouseEventPress);
//    QApplication::sendEvent(this, &mouseEventRelease);
   //  webView -> setFocus();
}

void WebView::keyPressEvent(QKeyEvent *event)
{
   if (event->key() == Qt::Key_Return && (event->modifiers() & Qt::AltModifier))
      emit editingFinishKeyPressed();
   else if (event->key() == Qt::Key_F12)
      QMessageBox::information(this, "", page() -> mainFrame() -> toHtml()
                               .replace("<", "&lt;").replace(">", "&gt;").replace("\n", "<br>"));
   else
      QWebView::keyPressEvent(event);
}

void WebView::dragEnterEvent(QDragEnterEvent *event)
{
   if (event->mimeData()->hasUrls())
      event->acceptProposedAction();   
}

void WebView::dropEvent(QDropEvent *event)
{
   emit dropMimeDataSignal(event -> mimeData());
}

//----------------------------------------------------------------------

TextEditorObject::TextEditorObject(TextEditor *_textEditor, const QString &_id)
   : textEditor(_textEditor)
{
   id = _id;
}

void TextEditorObject::setId(const QString &_id)
{
   id = _id;
}

QString TextEditorObject::getId() const
{
   return id;
}

//----------------------------------------------------------------------

TextEditorObjectImage::TextEditorObjectImage(TextEditor *_textEditor, const QString &_id, const QImage &image)
   : TextEditorObject(_textEditor, _id)
{
   setImage(image);
}

TextEditorObjectImage::TextEditorObjectImage(TextEditor *_textEditor, const QString &_id, const QByteArray &byteArray)
   : TextEditorObject(_textEditor, _id)
{
   setByteArray(byteArray);
}

TextEditorObjectImage::~TextEditorObjectImage()
{
   removeFile(filePath);
}

void TextEditorObjectImage::setImage(const QImage &image, bool replace)
{
   QBuffer buffer;
   QImageWriter writer(&buffer, "jpg");
   if (!writer.write(image))
      qDebug("TextEditorObjectImage::setImage(): could not write to buffer!");

   setByteArray(buffer.data(), replace);
}

QImage TextEditorObjectImage::getImage() const
{
   QImage image;
   if (!image.load(filePath))
      qDebug("TextEditorObjectImage::getImage(): could not load the image");
   return image;
}

QString TextEditorObjectImage::getType() const
{
   return ATTACHMENT_IMAGE;
}

QByteArray TextEditorObjectImage::getByteArray() const
{
   QFile file (filePath);
   if (!file.open(QIODevice::ReadOnly))
   {
      qDebug("TextEditorObjectImage::getByteArray(): could not open the file");
      return QByteArray();
   }

   return file.readAll();
}

void TextEditorObjectImage::setByteArray(const QByteArray &byteArray, bool replace)
{
   if (!filePath.isEmpty())
      removeFile(filePath);

   QString initialId = getId();
   if (replace)
      setId(Random::random() -> getARandomName());

#ifdef Q_WS_WIN
   filePath = QDir::toNativeSeparators(QDir::homePath()) + QDir::separator() + "flashqard";
   QDir().mkpath(filePath);
   filePath += QDir::separator() + getId();
#else
   filePath = QDir::tempPath() + QDir::separator() + getId();
#endif
   QFile::remove(filePath); //just in case it already exists
   QFile file (filePath);
   if (!file.open(QIODevice::WriteOnly))
      QMessageBox::warning(0, QObject::tr("Text Editor"),
                           QObject::tr("Unable to create a temporary file for embedded object: %1").arg(getId()));

   file.write(byteArray);
   

   textEditor -> evaluateJavaScript(QString("replaceId(\"%1\", \"%2\")").arg(initialId).arg(getId()), false);
   textEditor -> evaluateJavaScript("adjustImagesPath()", false);
   emit modifiedSignal();
}

QString TextEditorObjectImage::getFilePath() const
{
   return filePath;
}

void TextEditorObjectImage::removeFile(const QString &file)
{
   if (!QFile::remove(file))
      qDebug("TextEditorObjectImage::~TextEditorObjectImage(): could not remove file %s", file.toAscii().constData());
}

//----------------------------------------------------------------------

TextEditor::TextEditor(QWidget *parent)
   : QWidget(parent)
{
   objectsAreModified = false;
   readOnly = false;
   objForJavaScript = new ObjectsForJavaScript(this);
   connect(objForJavaScript, SIGNAL(clickedOnObjectSignal(const QString&)), this, SLOT(clickedOnObjectSlot(const QString&)));
   connect(objForJavaScript, SIGNAL(clickedOnNonObjectSignal()), this, SLOT(clickedOnNonObjectSlot()));
   webView = new WebView(this);
   connect(webView->page()->mainFrame(), SIGNAL(javaScriptWindowObjectCleared()), this, SLOT(addJavaScriptObjectSlot()));
   QFile file(":/TextEditorInitialHtml");
   if (!file.open(QIODevice::ReadOnly))
      QMessageBox::warning(0, QObject::tr("TextEditor"),
                           QObject::tr("TextEditor: Can not open template file."));
   else
      webView -> setHtml(QString(file.readAll()));

   connect(webView->page(), SIGNAL(selectionChanged()), this, SLOT(adjustActions()));
   connect(webView->page(), SIGNAL(selectionChanged()), this, SIGNAL(selectionChanged()));
   connect(webView->page(), SIGNAL(contentsChanged()), this, SLOT(somethingEdittedSlot()));
   connect(webView->page(), SIGNAL(contentsChanged()), this, SIGNAL(contentsChanged()));
   connect(webView, SIGNAL(focusOutSignal()), this, SLOT(finishEditing()));
   connect(webView, SIGNAL(editingFinishKeyPressed()), this, SIGNAL(editingFinishKeyPressed()));
   connect(webView, SIGNAL(dropMimeDataSignal(const QMimeData*)), this, SLOT(dropMimeDataSlot(const QMimeData*)));
//   webView->page()->action(QWebPage::Reload) -> setShortcut(Qt::Key_F5);
      
   createLayout();
   createActions();
   adjustActions();
   adjustFontSizes();
}

void TextEditor::setReadOnly(bool _readOnly)
{
   readOnly = _readOnly;
   evaluateJavaScript(QString("setReadOnly(%1)").arg(readOnly?"true":"false"), false);
}

void TextEditor::addJavaScriptObjectSlot()
{
   webView->page()->mainFrame()->addToJavaScriptWindowObject("objFromCpp", objForJavaScript);
}

void TextEditor::somethingEdittedSlot()
{
   adjustActions();
}

void TextEditor::updateStyle()
{
   evaluateJavaScript("updateCSS()", false);
   setZoomFactor(Settings::settings() -> cardAppearanceZoomFactor);
}

void TextEditor::dropMimeDataSlot(const QMimeData *mime)
{
   if (mime -> hasUrls())
   {
      QList<QUrl> urls = mime->urls();
      if (!urls.size())
         return;
      
      Downloader downloader(this);
      insertImage(downloader.download(urls[0]));
   }
}

void TextEditor::clickedOnObjectSlot(const QString &id)
{
   emit clickedOnObjectSignal(findObject(id));
}

void TextEditor::clickedOnNonObjectSlot()
{
   emit clickedOnNonObjectSignal();
}

void TextEditor::anObjectIsModifiedSlot()
{
   objectsAreModified = true;
}

void TextEditor::createLayout()
{
   QVBoxLayout *mainLayout = new QVBoxLayout(this);
   mainLayout -> addWidget(webView);
   
}

void TextEditor::createActions()
{
   alignLeftAction = new QAction(tr("Align left"), this);
   alignLeftAction -> setCheckable(true);
   connect(alignLeftAction, SIGNAL(triggered()), this, SLOT(alignLeftSlot()));

   alignRightAction = new QAction(tr("Align right"), this);
   alignRightAction -> setCheckable(true);
   connect(alignRightAction, SIGNAL(triggered()), this, SLOT(alignRightSlot()));
   
   alignCenterAction = new QAction(tr("Align center"), this);
   alignCenterAction -> setCheckable(true);
   connect(alignCenterAction, SIGNAL(triggered()), this, SLOT(alignCenterSlot()));
   
   alignJustifyAction = new QAction(tr("Align justify"), this);
   alignJustifyAction -> setCheckable(true);
   connect(alignJustifyAction, SIGNAL(triggered()), this, SLOT(alignJustifySlot()));

   strikeThroughAction = new QAction(tr("Strike through"), this);
   strikeThroughAction -> setCheckable(true);
   connect(strikeThroughAction, SIGNAL(triggered()), this, SLOT(strikeThroughSlot()));

   numberedListAction = new QAction(tr("Numbered list"), this);
   numberedListAction -> setCheckable(true);
   connect(numberedListAction, SIGNAL(triggered()), this, SLOT(numberedListSlot()));

   bulletedListAction = new QAction(tr("Bulleted list"), this);
   bulletedListAction -> setCheckable(true);
   connect(bulletedListAction, SIGNAL(triggered()), this, SLOT(bulletedListSlot()));
}

void TextEditor::setDocument(const Document &newDocument)
{
   document = newDocument;
   objectsAreModified = false;
   clearObjects();
   webView -> page() -> undoStack() -> clear();
   addDocumentAttachmentsToObjects();
   evaluateJavaScript(QString("replaceHtml(\"%1\")").arg(document.getText().replace("\"", "\\\"").simplified()), true);
   updateStyle();
}

const Document& TextEditor::getDocument()
{
   document.setText(getHtml());
   objectsAreModified = false;

   addUsedAttachmentsToDocument();
   //undo stack must get cleared otherwise deletion of an image can be
   //undone while the image doesn't really exist in the attachments of
   //the document.
   webView -> page() -> undoStack() -> clear();
   return document;
}

void TextEditor::reload()
{
   getAction(QWebPage::Reload) -> trigger();
}


void TextEditor::clear()
{
   setDocument(Document());
}

void TextEditor::startEditing()
{
   webView -> startEditing();
}

void TextEditor::finishEditing()
{
   QString currentHtml = getHtml();
   if ((document.getText() != currentHtml || objectsAreModified) && this->isEnabled() && !readOnly)
   {
      document.setText(currentHtml);
      emit editingFinished();
   }
}

bool TextEditor::insertImage(const QString &imageUrl)
{
   QByteArray byteArray = loadImage(imageUrl);
   return insertImage(byteArray);
}

bool TextEditor::insertImage(const QByteArray &byteArray)
{
   //first just test whether it is a valid image or not
   QImage image;
   if (!image.loadFromData(byteArray))
   {
      QMessageBox::warning(this, tr("Insert Image"),
                           tr("Inserting image failed."));
      return false;
   }

   //resize if it is bigger than 700x500
   if (image.size().width() > 700 || image.size().height() > 500)
      image = image.scaled(QSize(700, 500), Qt::KeepAspectRatio, Qt::SmoothTransformation);
   
   QString id = Random::random() -> getARandomName();
   TextEditorObjectImage *imageObject = new TextEditorObjectImage(this, id, image);
   connect(imageObject, SIGNAL(modifiedSignal()), this, SLOT(anObjectIsModifiedSlot()));
   objects << imageObject;
   QString path = imageObject -> getFilePath();
#ifdef Q_WS_WIN
   path.replace("\\", "\\\\");
#endif
   QString html = QString("<p><img id=\\\"%1\\\" src=\\\"%2\\\" /></p>").arg(id).arg(path);
   execCommand("insertHTML", html);

   //simulate a click to activate image editor (in advanced text editor)
   emit clickedOnObjectSignal(imageObject);
   return true;
}

QByteArray TextEditor::loadImage(const QString &imagePath)
{
   QFile file(imagePath);
   file.open(QIODevice::ReadOnly);
   return file.readAll();
}

void TextEditor::setFontFamily(const QString &family)
{
   execCommand("fontName", family);
}

QString TextEditor::getFontFamily() const
{
   return const_cast<TextEditor*>(this) -> evaluateJavaScript("getFontFamily()", false).toString();
}

void TextEditor::setFontSize(int fontSize)
{
   //setFontSize() function requires a number in the range 0-6
   //But the function below requires a number in the range 1-7
   execCommand("fontSize", QString::number(fontSize+1));
}

int TextEditor::getFontSize() const
{
   QString fontSizeString = const_cast<TextEditor*>(this) -> evaluateJavaScript("getFontSize()", false).toString();
   qDebug("TextEditor::getFontSize(): %s", fontSizeString.toAscii().constData());

   fontSizeString.chop(2);
   int fontSizeInt = fontSizeString.toInt();
   
   fontSizeInt = ceil(static_cast<double>(fontSizeInt)/getZoomFactor() - 0.49 );

   return fontSizes.indexOf(fontSizeInt);
}

void TextEditor::setForegroundColor(const QColor &color)
{
   execCommand("foreColor", color.name());
}

QColor TextEditor::getForegroundColor() const
{
   return rgbToColor(const_cast<TextEditor*>(this) -> evaluateJavaScript("getForegroundColor()", false).toString());
}

void TextEditor::setBackgroundColor(const QColor &color)
{
   execCommand("backColor", color.name());
}

QColor TextEditor::getBackgroundColor() const
{
   return rgbToColor(const_cast<TextEditor*>(this) -> evaluateJavaScript("getBackgroundColor()", false).toString());
}

void TextEditor::setPageBackgroundColor(const QColor &color)
{
   evaluateJavaScript(QString("setPageBackgroundColor(\"%1\")").arg(color.name()), false);
}

QColor TextEditor::getPageBackgroundColor() const
{
   return rgbToColor(const_cast<TextEditor*>(this) -> evaluateJavaScript("getPageBackgroundColor()", false).toString());
}

void TextEditor::setAlignment(Qt::Alignment alignment)
{
   if (alignment&Qt::AlignLeft)
      alignLeftSlot();
   else if (alignment&Qt::AlignRight)
      alignRightSlot();
   else if (alignment&Qt::AlignCenter)
      alignCenterSlot();
   else if (alignment&Qt::AlignJustify)
      alignJustifySlot();
}

void TextEditor::setZoomFactor(double zoom)
{
   webView -> setTextSizeMultiplier(zoom);
}

double TextEditor::getZoomFactor() const
{
   return webView -> textSizeMultiplier();
}

Qt::Alignment TextEditor::getAlignment() const
{
   QString alignment = const_cast<TextEditor*>(this) -> evaluateJavaScript("getAlignment()", false).toString();
   return alignment == "left" ? Qt::AlignLeft
      : alignment == "right" ? Qt::AlignRight
      : alignment == "center" ? Qt::AlignCenter
      : Qt::AlignJustify;
}

void TextEditor::execCommand(const QString &cmd)
{
//   qDebug("TextEditor::execCommand(%s)", cmd.toAscii().constData());
   QWebFrame *frame = webView->page()->mainFrame();
   QString js = QString("document.execCommand(\"%1\", false, null)").arg(cmd);
   frame->evaluateJavaScript(js);
}

void TextEditor::execCommand(const QString &cmd, const QString &arg)
{
//   qDebug("TextEditor::execCommand(%s, %s) began", cmd.toAscii().constData(), arg.toAscii().constData());
   QWebFrame *frame = webView->page()->mainFrame();
   QString js = QString("document.execCommand(\"%1\", false, \"%2\")").arg(cmd).arg(arg);
   frame->evaluateJavaScript(js);
}

bool TextEditor::queryCommandState(const QString &cmd)
{
   QWebFrame *frame = webView->page()->mainFrame();
   QString js = QString("document.queryCommandState(\"%1\", false, null)").arg(cmd);
   QVariant result = frame->evaluateJavaScript(js);
   return result.toString().simplified().toLower() == "true";
}

void TextEditor::adjustActions()
{
   Qt::Alignment alignment = getAlignment();
   alignLeftAction -> setChecked(alignment&Qt::AlignLeft);
   alignRightAction -> setChecked(alignment&Qt::AlignRight);
   alignCenterAction -> setChecked(alignment&Qt::AlignCenter);
   alignJustifyAction -> setChecked(alignment&Qt::AlignJustify);

   QString direction = evaluateJavaScript("getTextDirection()", false).toString();
   getAction(QWebPage::SetTextDirectionLeftToRight) -> setChecked(direction == "ltr");
   getAction(QWebPage::SetTextDirectionRightToLeft) -> setChecked(direction == "rtl");

   strikeThroughAction->setChecked(queryCommandState("strikeThrough"));
   numberedListAction->setChecked(queryCommandState("insertOrderedList"));
   bulletedListAction->setChecked(queryCommandState("insertUnorderedList"));
}

QAction* TextEditor::getAction(int action) const
{
   if (action >= 0 && action <= 52)
      return webView -> page() -> action(static_cast<QWebPage::WebAction>(action));

   if (action == AlignLeft)
      return alignLeftAction;
   else if (action == AlignRight)
      return alignRightAction;
   else if (action == AlignCenter)
      return alignCenterAction;
   else if (action == AlignJustify)
      return alignJustifyAction;
   else if (action == StrikeThrough)
      return strikeThroughAction;
   else if (action == NumberedList)
      return numberedListAction;
   else if (action == BulletedList)
      return bulletedListAction;


   qDebug("TextEditor::getAction() is returning 0");
   return 0;
}

void TextEditor::addCommand(QUndoCommand *command)
{
   webView -> page() -> undoStack() -> push(command);
}

void TextEditor::focusInEvent(QFocusEvent *event)
{
   QWidget::focusInEvent(event);
   webView -> setFocus();
}

void TextEditor::alignLeftSlot()
{
   execCommand("justifyLeft");
}

void TextEditor::alignCenterSlot()
{
   execCommand("justifyCenter");
}

void TextEditor::alignRightSlot()
{
   execCommand("justifyRight");
}

void TextEditor::alignJustifySlot()
{
   execCommand("justifyFull");
}

void TextEditor::strikeThroughSlot()
{
   execCommand("strikeThrough");
}

void TextEditor::numberedListSlot()
{
    execCommand("insertOrderedList");
}

void TextEditor::bulletedListSlot()
{
    execCommand("insertUnorderedList");
}

QVariant TextEditor::evaluateJavaScript(const QString &cmd, bool itIsEditting)
{
//   qDebug("TextEditor::evaluateJavaScript: %s", cmd.toAscii().constData());
   QVariant returnValue = webView -> page() -> mainFrame() -> evaluateJavaScript(cmd);
   if (itIsEditting)
      somethingEdittedSlot();
   return returnValue;
}

void TextEditor::adjustFontSizes()
{
   //we want to set font sizes like x-large xx-large,... but
   //the function getFontSize() which is called by javaScript always
   //returns in px like 42px.
   //so we create a table of values.
   //There are 7 values: xx-small, x-small, small, medium,
   //large, x-large, xx-large
   
   //NOTE: test it in other platforms...

   fontSizes << 9 << 11 << 14 << 17 << 21 << 28 << 42;
}

QColor TextEditor::rgbToColor(QString rgb) const
{
   rgb.chop(1);
   rgb.remove(0, 4);
   QStringList list = rgb.split(",");
   if (list.size() == 3)
      return QColor(list[0].toInt(), list[1].toInt(), list[2].toInt());
   else
      return QColor();
}

QString TextEditor::getHtml() const
{
   QString html = const_cast<TextEditor*>(this) -> evaluateJavaScript("getHtml()", false).toString();
   HtmlParser *htmlParser = HtmlParser::htmlParser();
   int iterator = 0;
   while (htmlParser->setTagAttribute(html, iterator, "img", "src", ""));
   return html;
}

// QString TextEditor::addObject(const QByteArray &byteArray, const QString &id)
// {
//    QFile file (QDir::tempPath() + QDir::separator() + id);
//    if (!file.open(QIODevice::WriteOnly))
//    {
//       QMessageBox::warning(0, QObject::tr("Text Editor"),
//                            QObject::tr("Unable to create a temporary file for embedded object: %1").arg(id));
//       return false;
//    }
//    file.write(byteArray);
//    QFileInfo info(file);
//    objects[id] = info.filePath();
   
//    return info.filePath();
// }

// QString TextEditor::getObjectPath(const QString &id) const
// {
//    return objects[id];
// }

void TextEditor::clearObjects()
{
   while (objects.size())
      delete objects.takeAt(0);
   
//    QStringList list = objects.values();
//    for (int i=0; i<list.size(); i++)
//       if (!QFile::remove(list[i]))
//          qDebug("TextEditor::clearObjects(): could not remove file: %s", list[i].toAscii().constData());
//    objects.clear();
}

void TextEditor::addDocumentAttachmentsToObjects()
{
   for (int i=0; i<document.attachmentCount(); i++)
   {
      if (document.getAttachmentType(i) == ATTACHMENT_IMAGE)
      {
         TextEditorObjectImage *imageObject = new TextEditorObjectImage(this, document.getAttachmentName(i),
                                                                        document.attachmentToByteArray(i));
         connect(imageObject, SIGNAL(modifiedSignal()), this, SLOT(anObjectIsModifiedSlot()));
         objects << imageObject;
      }
   }
}

// void TextEditor::removeUnusedAttachments()
// {
//    QStringList ids = evaluateJavaScript("getAllImagesId()", false).toString().split(",", QString::SkipEmptyParts);
   
//    for (int i=0; i<document.attachmentCount(); i++)
//       if (!ids.contains(document.getAttachmentName(i)))
//          document.deleteAttachment(i--);
// }

TextEditorObject* TextEditor::findObject(const QString &id)
{
   for (int i=0; i<objects.size(); i++)
      if (objects[i] -> getId() == id)
         return objects[i];
   
   qDebug("TextEditor::findObject(): returning 0");
   return 0;
}

void TextEditor::addUsedAttachmentsToDocument()
{
   document.deleteAllAttachments();

   QStringList usedIds = evaluateJavaScript("getAllImagesId()", false).toString().split(",", QString::SkipEmptyParts);

   for (int i=0; i<objects.size(); i++)
   {
      if (usedIds.contains(objects[i]->getId()))
      {
         document.addAttachment(objects[i]->getByteArray(), objects[i] -> getId(), objects[i] -> getType());
         qDebug("TextEditor::addUsedAttachmentsToDocument(): adding object %s", objects[i]->getId().toAscii().constData());
      }
   }
}


//--------------------------------------------------------------------------------

CommandResizeImage::CommandResizeImage(TextEditorObjectImage *_imageObject, const QSize &_size)
   : imageObject(_imageObject), newSize(_size)
{
   setText(QObject::tr("Resize image"));
   initialImage = imageObject -> getImage();
}

void CommandResizeImage::undo()
{
   imageObject -> setImage(initialImage, true);
}

void CommandResizeImage::redo()
{
   imageObject -> setImage(initialImage.scaled(newSize, Qt::IgnoreAspectRatio, Qt::SmoothTransformation), true);
}

//----------------------------------------------------------------------

CommandFlipImage::CommandFlipImage(TextEditorObjectImage *_imageObject, Qt::Orientation _orientation)
   : imageObject(_imageObject), orientation(_orientation)
{
   if (orientation == Qt::Vertical)
      setText(QObject::tr("Flip image"));
   else
      setText(QObject::tr("Mirror image"));
}

void CommandFlipImage::undo()
{
   redo();
}

void CommandFlipImage::redo()
{
   imageObject -> setImage(imageObject->getImage().mirrored(orientation==Qt::Horizontal, orientation==Qt::Vertical), true);
}

//----------------------------------------------------------------------

CommandRotateImage::CommandRotateImage(TextEditorObjectImage *_imageObject, Rotation _rotation)
   : imageObject(_imageObject), rotation(_rotation)
{
   if (rotation == RotateRight)
      setText(QObject::tr("Rotate image right"));
   else
      setText(QObject::tr("Rotate image left"));
}

void CommandRotateImage::undo()
{
   QMatrix matrix;
   matrix.rotate(rotation==RotateRight? -90 : 90);
   imageObject -> setImage(imageObject->getImage().transformed(matrix, Qt::SmoothTransformation), true);
}

void CommandRotateImage::redo()
{
   QMatrix matrix;
   matrix.rotate(rotation==RotateRight? 90 : -90);
   imageObject -> setImage(imageObject->getImage().transformed(matrix, Qt::SmoothTransformation), true);
}

// ChangeTextDirectionCommand::ChangeTextDirectionCommand(TextEditor *_textEditor, Qt::LayoutDirection _direction)
//    : textEditor(_textEditor), direction(_direction)
// {
//    setText(QObject::tr("Change text direction"));
// }

// void ChangeTextDirectionCommand::undo()
// {
//    qDebug("ChangeTextDirectionCommand::undo()");
//    textEditor -> evaluateJavaScript(QString("setTextDirection(\"%1\", \"%2\")")
//                                     .arg(direction==Qt::LeftToRight?"rtl":"ltr").arg(tagId), true);
// }

// void ChangeTextDirectionCommand::redo()
// {
//    qDebug("ChangeTextDirectionCommand::redo()");
//    tagId = textEditor -> evaluateJavaScript(QString("setTextDirection(\"%1\", \"%2\")")
//                                             .arg(direction==Qt::LeftToRight?"ltr":"rtl").arg(tagId), true).toString();
// }


//============================================================

PlainTextEditor::PlainTextEditor(QWidget *parent)
   :QTextEdit(parent)
{
   setAcceptRichText(false);
}

void PlainTextEditor::setText(const QString &newText)
{
   initialText = newText;
   setPlainText(newText);
}

void PlainTextEditor::clear()
{
   setText(QString());
}

QString PlainTextEditor::getText() const
{
   return toPlainText();
}

void PlainTextEditor::focusOutEvent(QFocusEvent *event)
{
   QTextEdit::focusOutEvent(event);
   if (toPlainText() != initialText)
   {
      initialText = toPlainText();
      emit editingFinished();
   }
}

void PlainTextEditor::keyPressEvent(QKeyEvent *event)
{
   if (event->key() == Qt::Key_Return && (event->modifiers() & Qt::AltModifier))
      emit editingFinishKeyPressed();
   else
      QTextEdit::keyPressEvent(event);
}

bool PlainTextEditor::canInsertFromMimeData(const QMimeData *source) const
{
   if (source->hasHtml())
      return true;
   else
      return QTextEdit::canInsertFromMimeData(source);
}

void PlainTextEditor::insertFromMimeData(const QMimeData *source)
{
   if (source->hasHtml())
   {
      QString plainText = HtmlParser::htmlParser()->htmlToPlainText(source->html());
      QTextCursor cursor = textCursor();
      cursor.insertText(plainText);
   }
   else
      QTextEdit::insertFromMimeData(source);

   this -> setFocus();
}

//======================================================================

LineEditor::LineEditor(QWidget *parent)
   : QLineEdit(parent)
{

}

void LineEditor::setText(const QString &newText)
{
   initialText = newText;
   QLineEdit::setText(newText);
}

QString LineEditor::getText() const
{
   return text();
}

void LineEditor::clear()
{
   setText(QString());
}

void LineEditor::focusOutEvent(QFocusEvent *event)
{
   QLineEdit::focusOutEvent(event);
   if (text() != initialText)
   {
      initialText = text();
      emit editingFinished();
   }
}

void LineEditor::keyPressEvent(QKeyEvent *event)
{
   if (event->key() == Qt::Key_Return && (event->modifiers() & Qt::AltModifier))
      emit editingFinishKeyPressed();
   else
      QLineEdit::keyPressEvent(event);
}

